from aioinject.container import Container, SyncContainer
from aioinject.context import Context, SyncContext
from aioinject.decorators import INJECTED, Inject, Injected
from aioinject.providers import Provider
from aioinject.providers.context import FromContext
from aioinject.providers.object import Object
from aioinject.providers.scoped import Scoped, Singleton, Transient
from aioinject.scope import Scope


__all__ = [
    "INJECTED",
    "Container",
    "Context",
    "FromContext",
    "Inject",
    "Injected",
    "Injected",
    "Object",
    "Provider",
    "Scope",
    "Scoped",
    "Singleton",
    "SyncContainer",
    "SyncContext",
    "Transient",
]



import logging
from typing import TypeVar

from aioinject import Context
from aioinject.context import ProviderRecord
from aioinject.extensions import OnResolveExtension


T = TypeVar("T")


logger = logging.getLogger(__name__)


class MyExtension(OnResolveExtension):
    async def on_resolve(
        self,
        context: Context,  # noqa: ARG002
        provider: ProviderRecord[T],
        instance: T,  # noqa: ARG002
    ) -> None:
        logger.info("%s type was provided!", provider.info.type_)



import httpx
import pytest
from _pytest.fixtures import SubRequest
from starlette.testclient import TestClient


@pytest.fixture(
    params=[
        "/async-route",
        "/async-depends",
        "/asyncgen-depends",
    ]
)
async def route(request: SubRequest) -> str:
    return request.param


async def test_function_route(
    http_client: httpx.AsyncClient,
    provided_value: int,
    route: str,
) -> None:
    response = await http_client.get(route)
    assert response.status_code == httpx.codes.OK.value
    assert response.json() == {"value": provided_value}


async def test_can_inject_context(
    http_client: httpx.AsyncClient,
) -> None:
    """Should be able to inject Request and BackgroundTasks from context"""
    response = await http_client.get("/context-injection")
    assert response.status_code == httpx.codes.OK.value
    assert response.json() is True


async def test_websocket_route(
    starlette_http_client: TestClient,
    provided_value: int,
) -> None:
    with starlette_http_client.websocket_connect("/ws") as websocket:
        result = websocket.receive_json()
        assert result == {"value": provided_value}



import dataclasses

import pytest

from aioinject import Container, Scoped


class _A:
    pass


@dataclasses.dataclass
class _B:
    a: _A


@dataclasses.dataclass
class _C:
    b: _B


@pytest.fixture
def container() -> Container:
    container = Container()
    container.register(Scoped(_A))
    container.register(Scoped(_B))
    container.register(Scoped(_C))
    return container



import random
import uuid
from typing import TypedDict

import pytest

import aioinject


@pytest.fixture
def provided_value() -> int:
    return random.randint(1, 1_000_000)


class NumberService:
    def __init__(self, value: int) -> None:
        self._value = value

    def get_number(self) -> int:
        return self._value


class ScopedNode(TypedDict):
    """A node with unique id per scope."""

    id: str


def get_node() -> ScopedNode:
    return {"id": uuid.uuid4().hex}


@pytest.fixture
def container(provided_value: int) -> aioinject.Container:
    container = aioinject.Container()
    container.register(aioinject.Object(provided_value))
    container.register(aioinject.Scoped(NumberService))
    container.register(aioinject.Scoped(get_node))
    return container



from __future__ import annotations

from typing import Generic

from aioinject import Container, Scoped
from aioinject._types import T


class TestGeneric(Generic[T]):
    pass


async def dependency() -> TestGeneric[int]:
    return TestGeneric()


async def dependant(_: TestGeneric[int]) -> int:
    return 42


async def test_provide_parametrized_generic() -> None:
    container = Container()
    container.register(Scoped(dependency), Scoped(dependant))
    async with container.context() as ctx:
        await ctx.resolve(int)



from __future__ import annotations

import abc
from collections.abc import Callable
from typing import Any

from django.http import HttpRequest, HttpResponse

from aioinject import SyncContainer, SyncContext
from aioinject._types import P, T
from aioinject.decorators import base_inject


try:
    import rest_framework.request
except ImportError:  # pragma: no cover
    rest_framework = None  # type: ignore[assignment]


def context_getter(
    args: tuple[Any, ...], kwargs: dict[str, Any]
) -> SyncContext:
    request: HttpRequest | rest_framework.request.Request | None = next(
        (
            # Django/DRF always seem to pass request through args
            arg
            for arg in args
            if isinstance(
                arg,
                HttpRequest
                if rest_framework is None
                else (HttpRequest, rest_framework.request.Request),
            )
        ),
        None,
    )
    if request is None:  # pragma: no cover
        msg = (
            f"Could not find request parameter, expected one of the parameters to be 'django.http.HttpRequest' or 'rest_framework.request.Request':\n"
            f"  {args=}, {kwargs=}"
        )
        raise ValueError(msg)
    return request.__aioinject_context__  # type: ignore[attr-defined]


def inject(function: Callable[P, T]) -> Callable[P, T]:
    return base_inject(
        function,
        context_parameters=(),
        context_getter=context_getter,
    )


class SyncAioinjectMiddleware(abc.ABC):
    def __init__(
        self, get_response: Callable[[HttpRequest], HttpResponse]
    ) -> None:
        self.get_response = get_response

    @property
    @abc.abstractmethod
    def container(self) -> SyncContainer:
        raise NotImplementedError

    def __call__(self, request: HttpRequest) -> HttpResponse:
        with self.container.context() as ctx:
            request.__aioinject_context__ = ctx  # type: ignore[attr-defined]
            return self.get_response(request)



import inspect
from collections.abc import Awaitable, Callable
from typing import Any

from aiogram import BaseMiddleware, Router
from aiogram.types import TelegramObject

import aioinject
from aioinject._types import P, T
from aioinject.decorators import add_parameters_to_signature, base_inject


__all__ = ["AioInjectMiddleware", "inject"]


_ARG_NAME = "aioinject_context"


class _ContextGetter:
    def __init__(self, *, remove_from_args: bool) -> None:
        self.remove_from_args = remove_from_args

    def __call__(
        self, args: tuple[Any], kwargs: dict[str, Any]
    ) -> aioinject.Context:
        if _ARG_NAME in kwargs:  # pragma: no cover
            if self.remove_from_args:
                return kwargs.pop(_ARG_NAME)
            return kwargs[_ARG_NAME]

        # Try to find a dict-like looking object (in case it's a middleware)
        for arg in args:
            if not isinstance(arg, dict):
                continue
            if _ARG_NAME in arg:
                return arg[_ARG_NAME]

        msg = "Could not find aioinject context, did you forget to add AioinjectMiddleware?"
        raise ValueError(msg)


def inject(function: Callable[P, T]) -> Callable[P, T]:  # pragma: no cover
    signature = inspect.signature(function)
    existing_parameter = signature.parameters.get(_ARG_NAME)
    if not existing_parameter:
        add_parameters_to_signature(function, {_ARG_NAME: aioinject.Context})

    return base_inject(
        function,
        context_parameters=(),
        context_getter=_ContextGetter(remove_from_args=not existing_parameter),
    )


class AioInjectMiddleware(BaseMiddleware):
    def __init__(self, container: aioinject.Container) -> None:
        self.container = container

    async def __call__(
        self,
        handler: Callable[[TelegramObject, dict[str, Any]], Awaitable[Any]],
        event: TelegramObject,
        data: dict[str, Any],
    ) -> Any:
        async with self.container.context() as ctx:
            data[_ARG_NAME] = ctx
            return await handler(event, data)

    def add_to_router(self, router: Router) -> None:
        for observer in router.observers.values():
            observer.outer_middleware.register(middleware=self)



import functools
import inspect
from collections.abc import Mapping
from typing import Any

from aioinject._types import FactoryResult, FactoryType, T, _guess_return_type
from aioinject.dependencies import collect_parameters
from aioinject.errors import CannotDetermineReturnTypeError
from aioinject.extensions import ProviderExtension
from aioinject.extensions.providers import (
    CacheDirective,
    LockDirective,
    ProviderInfo,
    ResolveDirective,
)
from aioinject.providers import Provider
from aioinject.scope import BaseScope, Scope


class Scoped(Provider[T]):
    cache_ok: bool = True

    def __init__(
        self,
        factory: FactoryType[T],
        interface: type[T] | None = None,
        scope: BaseScope | None = None,
    ) -> None:
        self.implementation = factory
        self.interface = interface
        self.scope = scope

    def provide(self, kwargs: Mapping[str, Any]) -> FactoryResult[T]:
        return self.implementation(**kwargs)

    def __repr__(self) -> str:  # pragma: no cover
        return f"{self.__class__.__name__}({self.implementation=}, {self.interface=})"

    @functools.cached_property
    def is_async(self) -> bool:
        unwrapped = inspect.unwrap(self.implementation)
        return inspect.iscoroutinefunction(
            unwrapped
        ) or inspect.isasyncgenfunction(unwrapped)

    @functools.cached_property
    def is_context_manager(self) -> bool:
        unwrapped = inspect.unwrap(self.implementation)
        return inspect.isgeneratorfunction(
            unwrapped
        ) or inspect.isasyncgenfunction(unwrapped)


class Singleton(Scoped[T]):
    pass


class Transient(Scoped[T]):
    cache_ok = False


class ScopedProviderExtension(ProviderExtension[Scoped[Any]]):
    def __init__(
        self,
        default_singleton_scope: BaseScope = Scope.lifetime,
        default_scope: BaseScope = Scope.request,
    ) -> None:
        self._default_singleton_scope = default_singleton_scope
        self._default_scope = default_scope

    def supports_provider(self, provider: Scoped[object]) -> bool:
        return isinstance(provider, Scoped)

    def extract(
        self,
        provider: Scoped[T],
        type_context: Mapping[str, type[object]],
    ) -> ProviderInfo[T]:
        scope = (
            provider.scope
            if provider.scope is not None
            else (
                self._default_singleton_scope
                if isinstance(provider, Singleton)
                else self._default_scope
            )
        )

        dependencies = tuple(
            collect_parameters(
                dependant=provider.implementation,
                type_context=type_context,
            )
        )

        try:
            actual_type = _guess_return_type(
                provider.implementation, type_context=type_context
            )
        except CannotDetermineReturnTypeError:
            if not provider.interface:
                raise
            actual_type = provider.interface

        return ProviderInfo(
            interface=provider.interface or actual_type,
            type_=actual_type,
            dependencies=dependencies,
            scope=scope,
            compilation_directives=(
                CacheDirective(is_enabled=provider.cache_ok),
                ResolveDirective(
                    is_async=provider.is_async,
                    is_context_manager=provider.is_context_manager,
                ),
                LockDirective(is_enabled=isinstance(provider, Singleton)),
            ),
        )



from django.http import HttpRequest, HttpResponse

from aioinject import Injected
from aioinject.ext.django import inject


@inject
def function_view(_: HttpRequest, number: Injected[int]) -> HttpResponse:
    return HttpResponse(content=f"Number {number}")


@inject
def function_view_with_parameter(
    _: HttpRequest, parameter: str, number: Injected[int]
) -> HttpResponse:
    return HttpResponse(content=f"Number {parameter} {number}")



from __future__ import annotations

import textwrap


class Indent:
    def __init__(self, char: str = " " * 4, indent: int = 0) -> None:
        self._char = char
        self.indent = indent

    def format(self, text: str) -> str:
        return textwrap.indent(text, self._char * self.indent)



import dataclasses
from typing import Generic

from aioinject._types import T
from aioinject.scope import BaseScope


@dataclasses.dataclass(slots=True, kw_only=True)
class Dependency(Generic[T]):
    name: str
    type_: type[T]


@dataclasses.dataclass(slots=True, kw_only=True)
class CompilationDirective:
    is_enabled: bool = True


@dataclasses.dataclass(slots=True, kw_only=True)
class CacheDirective(CompilationDirective):
    optional: bool = True


@dataclasses.dataclass(slots=True, kw_only=True)
class ResolveDirective(CompilationDirective):
    is_async: bool
    is_context_manager: bool


@dataclasses.dataclass(slots=True, kw_only=True)
class LockDirective(CompilationDirective):
    pass


@dataclasses.dataclass(kw_only=True)
class ProviderInfo(Generic[T]):
    interface: type[T]
    type_: type[T]
    dependencies: tuple[Dependency[object], ...]
    scope: BaseScope
    compilation_directives: tuple[CompilationDirective, ...]



import pytest

from aioinject import Container, Scoped, Singleton
from aioinject.validation.errors import ValidationError
from aioinject.validation.rules import ScopeMismatchRule
from aioinject.validation.validate import validate_or_err


async def _make_str() -> str:
    return "42"


async def _make_int(s: str) -> int:
    return int(s)


def test_scope_mismatch() -> None:
    container = Container()
    container.register(Scoped(_make_str))
    container.register(Singleton(_make_int))

    with pytest.raises(ValidationError) as err_info:
        validate_or_err(container, (ScopeMismatchRule(),))

    assert str(err_info.value) == (
        "\n"
        "  scope-mismatch:\n"
        "    Provider(int) with scope Scope.lifetime depends on Provider(str) with scope Scope.request, which is lower"
    )



from __future__ import annotations

import asyncio
import contextlib
import dataclasses
import threading
from collections.abc import Callable
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from types import TracebackType
from typing import TYPE_CHECKING, Final, Generic

from typing_extensions import Self

from aioinject.scope import BaseScope, next_scope


if TYPE_CHECKING:
    from aioinject import Container, Provider, SyncContainer
    from aioinject.extensions import ProviderExtension
    from aioinject.extensions.providers import ProviderInfo

from aioinject._types import ExecutionContext, T


__all__ = ["Context", "ProviderRecord", "SyncContext"]


@dataclasses.dataclass(slots=True, kw_only=True)
class ProviderRecord(Generic[T]):
    provider: Provider[T]
    info: ProviderInfo[T]
    ext: ProviderExtension[Provider[T]]


class Context:
    def __init__(
        self,
        scope: BaseScope,
        context: ExecutionContext,
        container: Container,
        cache: dict[type[object], object] | None = None,
        lock_factory: Callable[
            [], AbstractAsyncContextManager[object]
        ] = asyncio.Lock,
    ) -> None:
        self.scope: Final = scope
        self.container: Final = container

        self._context = context.copy()
        self._context[scope] = self

        self.cache: dict[type[object], object] = (
            cache if cache is not None else {}
        )
        self.cache[type(self)] = self
        self.exit_stack = contextlib.AsyncExitStack()
        self.lock = lock_factory()

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.exit_stack.__aexit__(exc_type, exc_val, exc_tb)

    async def resolve(self, /, type_: type[T]) -> T:
        return await self.container.registry.compile(type_, is_async=True)(
            self._context,
            self.scope,
        )

    def context(
        self,
        context: dict[type[object], object] | None = None,
    ) -> Context:
        return Context(
            context=self._context,
            scope=next_scope(self.container.scopes, self.scope),
            container=self.container,
            cache=context,
        )

    def add_context(
        self,
        context: dict[type[object], object],
    ) -> None:
        for key, value in context.items():
            self.cache[key] = value


class SyncContext:
    def __init__(
        self,
        scope: BaseScope,
        context: ExecutionContext,
        container: SyncContainer,
        cache: dict[type[object], object] | None = None,
        lock_factory: Callable[
            [], AbstractContextManager[object]
        ] = threading.Lock,
    ) -> None:
        self.scope: Final = scope
        self.container: Final = container

        self._context = context.copy()
        self._context[scope] = self

        self.cache: dict[type[object], object] = (
            cache if cache is not None else {}
        )
        self.cache[type(self)] = self
        self.exit_stack = contextlib.ExitStack()
        self.lock = lock_factory()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.exit_stack.__exit__(exc_type, exc_val, exc_tb)

    def resolve(self, /, type_: type[T]) -> T:
        return self.container.registry.compile(type_, is_async=False)(
            self._context,
            self.scope,
        )

    def context(
        self,
        context: dict[type[object], object] | None = None,
    ) -> SyncContext:
        return SyncContext(
            context=self._context,
            scope=next_scope(self.container.scopes, self.scope),
            container=self.container,
            cache=context,
        )

    def add_context(
        self,
        context: dict[type[object], object],
    ) -> None:  # pragma: no cover
        for key, value in context.items():
            self.cache[key] = value



import collections
import typing

import pytest

from aioinject import (
    Container,
    Context,
    Object,
    Provider,
    Scoped,
    Singleton,
    Transient,
)
from aioinject._types import T
from aioinject.context import ProviderRecord
from aioinject.extensions import OnResolveExtension


@typing.final
class _TestExtension(OnResolveExtension):
    def __init__(self) -> None:
        self.type_counter: dict[type[object], int] = collections.defaultdict(
            int,
        )

    async def on_resolve(
        self,
        context: Context,  # noqa: ARG002
        provider: ProviderRecord[T],
        instance: T,  # noqa: ARG002
    ) -> None:
        self.type_counter[provider.info.type_] += 1


@pytest.mark.parametrize(
    "provider",
    [
        Object(0),
        Scoped(int),
        Transient(int),
        Singleton(int),
    ],
)
async def test_on_resolve(provider: Provider[int]) -> None:
    extension = _TestExtension()
    container = Container(extensions=(extension,))
    container.register(provider)

    async with container, container.context() as ctx:
        for i in range(1, 10 + 1):
            number = await ctx.resolve(int)
            assert number == 0
            assert extension.type_counter[int] == (
                i if isinstance(provider, Transient) else 1
            )



from collections.abc import Mapping
from typing import Any, final

import pytest
from pydantic_settings import BaseSettings

from aioinject import Container, Context, Scoped
from aioinject._types import FactoryResult, T
from aioinject.errors import ProviderNotFoundError
from aioinject.providers import Provider
from aioinject.providers.object import Object
from tests.testservices import AbstractImplA, AbstractImplB


@pytest.fixture
def container() -> Container:
    return Container()


def test_can_init(container: Container) -> None:
    assert container


def container_provider_mapping(
    container: Container,
) -> Mapping[type[object], list[Provider[object]]]:
    return {
        type_: [provider.provider for provider in providers]
        for type_, providers in container.registry.providers.items()
    }


def test_can_retrieve_context(container: Container) -> None:
    ctx = container.context()
    assert isinstance(ctx, Context)


def test_can_register_single(container: Container) -> None:
    provider = Scoped(AbstractImplA)
    container.register(provider)

    record = container.registry.providers[AbstractImplA][0]
    assert record.provider is provider
    assert record.info.interface is AbstractImplA
    assert record.info.type_ is AbstractImplA


def test_can_register_batch(container: Container) -> None:
    provider1 = Scoped(AbstractImplA)
    provider2 = Scoped(AbstractImplB)
    container.register(provider1, provider2)
    excepted = {AbstractImplA: [provider1], AbstractImplB: [provider2]}

    assert container_provider_mapping(container) == excepted


def test_register_unhashable_implementation(
    container: Container,
) -> None:
    class ExampleSettings(BaseSettings):
        value: list[str] = []

    container.register(Object([], interface=list[int]))
    container.register(Object(ExampleSettings(), interface=ExampleSettings))


def test_cant_register_multiple_providers_for_same_type(
    container: Container,
) -> None:
    container.register(Scoped(int))

    with pytest.raises(
        ValueError,
        match="^Provider for type <class 'int'> with same implementation already registered$",
    ):
        container.register(Scoped(int))


def test_can_retrieve_single_provider(container: Container) -> None:
    int_provider = Scoped(int)
    container.register(int_provider)
    assert container.registry.get_provider(int).provider is int_provider


def test_can_retrieve_multiple_providers(container: Container) -> None:
    int_providers = [
        Scoped(lambda: 1, int),
        Scoped(lambda: 2, int),
    ]
    container.register(*int_providers)
    assert len(container.registry.get_providers(int)) == len(int_providers)


def test_missing_provider() -> None:
    container = Container()
    with pytest.raises(ProviderNotFoundError) as exc_info:
        assert container.registry.get_provider(AbstractImplA)

    msg = f"Providers for type {AbstractImplA.__qualname__} not found"
    assert str(exc_info.value) == msg


def test_no_provider_extension() -> None:
    @final
    class CustomProvider(Provider[T]):
        implementation = int

        def provide(self, kwargs: dict[str, Any]) -> FactoryResult[T]:
            raise NotImplementedError

        def __repr__(self) -> str:
            return f"{self.__class__.__name__}()"

    container = Container()
    with pytest.raises(ValueError) as exc_info:  # noqa: PT011
        container.register(CustomProvider())

    assert (
        str(exc_info.value)
        == "ProviderExtension for provider CustomProvider() not found."
    )



import contextlib
from collections.abc import Iterator

from aioinject import SyncContainer
from aioinject.providers.scoped import Singleton


def test_should_close_singletons_sync() -> None:
    shutdown = False

    @contextlib.contextmanager
    def dependency() -> Iterator[int]:
        nonlocal shutdown
        yield 42
        shutdown = True

    container = SyncContainer()
    container.register(Singleton(dependency))
    with container:
        for _ in range(2):
            with container.context() as ctx:
                assert ctx.resolve(int) == 42  #  noqa: PLR2004

        assert shutdown is False
    assert shutdown is True



from typing import cast

import pytest
from _pytest.fixtures import SubRequest


pytest_plugins = [
    "anyio",
]


@pytest.fixture(scope="session", autouse=True, params=["asyncio", "trio"])
def anyio_backend(request: SubRequest) -> str:
    return cast("str", request.param)



import asyncio
import contextlib
from collections.abc import AsyncIterator

from aioinject import Container
from aioinject.extensions import LifespanExtension


class MyLifespanExtension(LifespanExtension):
    @contextlib.asynccontextmanager
    async def lifespan(
        self,
        container: Container,  # noqa: ARG002
    ) -> AsyncIterator[None]:
        print("Enter")
        yield None
        print("Exit")


async def main() -> None:
    container = Container(extensions=[MyLifespanExtension()])
    async with container:
        # print("Enter") is executed.
        pass
        # print("Exit") is executed.


if __name__ == "__main__":
    asyncio.run(main())



import uuid
from collections.abc import Callable
from http import HTTPStatus

import httpx
import pytest

from aioinject import (
    Container,
    Object,
    Provider,
    Scoped,
    Singleton,
    Transient,
)
from tests.integrations.utils import ExceptionPropagation, PropagatedError


@pytest.mark.parametrize(
    ("provider_type", "should_propagate"),
    [
        (Singleton, False),
        (Scoped, True),
        (Transient, True),
    ],
)
async def test_error_propagation(
    http_client: httpx.AsyncClient,
    container: Container,
    provider_type: Callable[..., Provider[object]],
    should_propagate: bool,
) -> None:
    propagation = ExceptionPropagation()

    container.register(provider_type(propagation.dependency))

    assert propagation.exc is None

    response = await http_client.get("/exception-propagation")
    assert response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR

    if should_propagate:
        assert isinstance(propagation.exc, PropagatedError)
    else:
        assert propagation.exc is None


async def test_ok(
    http_client: httpx.AsyncClient,
    container: Container,
) -> None:
    value = str(uuid.uuid4())
    container.register(Object(value))

    response = await http_client.get("/str")
    assert response.text == value


@pytest.mark.parametrize(
    "path",
    [
        "/request-no-context",
        "/request-context",
    ],
)
async def test_request_context(
    http_client: httpx.AsyncClient,
    path: str,
) -> None:
    value = str(uuid.uuid4())

    response = await http_client.post(path, content=value)
    assert response.text == value



from datetime import datetime
from typing import NewType

from aioinject import Container, SyncContainer, Transient
from aioinject.extensions import OnInitExtension


Now = NewType("Now", datetime)


class TimeExtension(OnInitExtension):
    def on_init(
        self,
        container: Container | SyncContainer,
    ) -> None:
        container.register(Transient(datetime.now, Now))


container = SyncContainer(extensions=[TimeExtension()])
with container.context() as ctx:
    print(ctx.resolve(Now))



import types
import typing
from collections.abc import Iterator, Sequence
from typing import Any

from aioinject._types import is_iterable_generic_collection
from aioinject.container import Container, Registry, SyncContainer
from aioinject.context import ProviderRecord
from aioinject.errors import ProviderNotFoundError
from aioinject.extensions.providers import ResolveDirective
from aioinject.scope import CurrentScope
from aioinject.validation.abc import ValidationRule
from aioinject.validation.errors import RuleViolation


def _iter_providers(registry: Registry) -> Iterator[ProviderRecord[Any]]:
    for provider_group in registry.providers.values():
        yield from provider_group


class NoAsyncDependenciesInSyncContainerRule(ValidationRule):
    def validate(
        self, container: Container | SyncContainer
    ) -> Sequence[RuleViolation]:
        if not isinstance(container, SyncContainer):
            return ()

        errors = []
        for provider in _iter_providers(container.registry):
            provide_directive = next(
                (
                    directive
                    for directive in provider.info.compilation_directives
                    if isinstance(directive, ResolveDirective)
                ),
                None,
            )
            if provide_directive is None:
                continue  # pragma: no cover

            if provide_directive.is_async:
                msg = f"Provider({provider.info.type_.__name__}) is async"
                errors.append(
                    RuleViolation(code="async-dependency", message=msg)
                )
        return errors


class ScopeMismatchRule(ValidationRule):
    def validate(  # noqa: C901
        self, container: Container | SyncContainer
    ) -> Sequence[RuleViolation]:
        errors = []
        for provider in _iter_providers(container.registry):
            for dependency in provider.info.dependencies:
                if not is_iterable_generic_collection(
                    dependency.type_
                ) and isinstance(dependency.type_, types.GenericAlias):
                    continue  # type: ignore[unreachable]

                try:
                    dependency_providers = container.registry.get_providers(
                        dependency.type_
                    )
                except ProviderNotFoundError:
                    if not is_iterable_generic_collection(dependency.type_):
                        raise  # pragma: no cover
                    dependency_providers = container.registry.get_providers(
                        typing.get_args(dependency.type_)[0]
                    )

                for dependency_provider in dependency_providers:
                    violation = self.validate_dependency(
                        container, provider, dependency_provider
                    )
                    if violation:
                        errors.append(violation)

        return errors

    def validate_dependency(
        self,
        container: Container | SyncContainer,
        provider: ProviderRecord[Any],
        dependency_provider: ProviderRecord[Any],
    ) -> RuleViolation | None:
        if isinstance(dependency_provider.info.scope, CurrentScope):
            return None

        scopes = tuple(container.scopes)
        scope_index = scopes.index(provider.info.scope)
        dependency_scope_index = scopes.index(dependency_provider.info.scope)

        if dependency_scope_index <= scope_index:
            return None

        msg = (
            f"Provider({provider.info.type_.__name__}) with scope {provider.info.scope} depends on "
            f"Provider({dependency_provider.info.type_.__name__}) with scope {dependency_provider.info.scope}, which is lower"
        )
        return RuleViolation(code="scope-mismatch", message=msg)


DEFAULT_RULES = [
    NoAsyncDependenciesInSyncContainerRule(),
    ScopeMismatchRule(),
]



import asyncio

from benchmark.benchmarks.context import context_benchmarks
from benchmark.benchmarks.simple import simple_benchmarks
from benchmark.lib.bench import Benchmark
from benchmark.lib.format import print_markdown_table


async def main() -> None:
    for collection in (simple_benchmarks, context_benchmarks):
        bench = Benchmark(
            benchmarks=collection.benchmarks,
        )
        results = await bench.run(rounds=[100_000])
        results = sorted(results, key=lambda result: result.mean)
        print_markdown_table(results)


if __name__ == "__main__":
    asyncio.run(main())



from typing import Generic, TypeVar

from aioinject import Container, Object, Scoped


T = TypeVar("T")
K = TypeVar("K")


class A(Generic[T]):
    def __init__(self, t: T) -> None:
        self.t = t


class B(Generic[T, K]):
    def __init__(self, t: T, k: K) -> None:
        self.t = t
        self.k = k


class BSpecialized(B[K, int]):
    pass


class Nested(Generic[T]):
    def __init__(self, obj: BSpecialized[T]) -> None:
        self.obj = obj


async def test_unbound_generic() -> None:
    container = Container()
    container.register(Scoped(A))
    container.register(Object(42))
    container.register(Object("string"))

    async with container.context() as ctx:
        a_int = await ctx.resolve(A[int])
        a_str = await ctx.resolve(A[str])

        assert a_int is not a_str  # type: ignore[comparison-overlap]

        assert a_int.t == 42  # noqa: PLR2004
        assert a_str.t == "string"


async def test_override() -> None:
    container = Container()

    container.register(Scoped(lambda: A(0), interface=A[int]))

    container.register(Scoped(A))
    container.register(Object(42))
    container.register(Object("string"))

    async with container.context() as ctx:
        a = await ctx.resolve(A[int])
        assert a.t == 0

        a_str = await ctx.resolve(A[str])
        assert a_str.t == "string"


async def test_specialized_generic() -> None:
    container = Container()

    container.register(Scoped(BSpecialized))
    container.register(Scoped(str))
    container.register(Scoped(int))

    async with container.context() as ctx:
        obj = await ctx.resolve(BSpecialized[str])
        assert obj.t == ""
        assert obj.k == 0


async def test_nested() -> None:
    container = Container()

    container.register(Scoped(Nested))
    container.register(Scoped(BSpecialized))
    container.register(Scoped(str))
    container.register(Scoped(int))

    async with container.context() as ctx:
        obj = await ctx.resolve(Nested[str])
        assert obj.obj.t == ""
        assert obj.obj.k == 0



from aioinject import Container, Object


async def test_union() -> None:
    container = Container()
    container.register(Object(42, interface=int | str))  # type: ignore[arg-type]

    async with container.context() as ctx:
        assert await ctx.resolve(int | str) == 42  # type: ignore[arg-type]   # noqa: PLR2004


async def test_union_generic_arg() -> None:
    container = Container()
    container.register(Object([], interface=list[int | str]))

    async with container.context() as ctx:
        assert await ctx.resolve(list[int | str]) == []



from aioinject import Scoped, Singleton, SyncContainer


class Database:
    def __init__(self) -> None:
        self._storage = {1: "Username"}

    def get(self, id: int) -> str | None:
        return self._storage.get(id)


class UserService:
    def __init__(
        self,
        database: Database,  # <- `Database` is injected here
    ) -> None:
        self._database = database

    def get(self, id: int) -> str:
        user = self._database.get(id)
        if user is None:
            raise ValueError
        return user


container = SyncContainer()
container.register(
    Singleton(Database),
    Scoped(UserService),
)

with (
    container,  # Singletons are managed
    container.context() as context,
):
    service = context.resolve(UserService)
    user = service.get(1)
    print(user)  # "Username"



import random
from http import HTTPStatus

from django.test import Client


def test_function_route(django_test_client: Client) -> None:
    response = django_test_client.get("/drf/api-view")
    assert response.status_code == HTTPStatus.OK
    assert response.json() == {"value": 42}


def test_api_view(django_test_client: Client) -> None:
    pk = random.randint(1, 1000)
    response = django_test_client.get(f"/drf/viewset/{pk}/")
    assert response.status_code == HTTPStatus.OK
    assert response.json() == {"value": 42, "id": str(pk)}



from __future__ import annotations

import contextlib
import dataclasses
import linecache
import typing
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, TypeVar

from aioinject._compilation.naming import (
    create_var_name,
    generate_factory_kwargs,
)
from aioinject._compilation.resolve import (
    AnyNode,
    FromContextNode,
    IterableNode,
    ProviderNode,
)
from aioinject._compilation.util import Indent
from aioinject._types import CompiledFn
from aioinject.errors import ScopeNotFoundError
from aioinject.extensions.providers import (
    CacheDirective,
    CompilationDirective,
    LockDirective,
    ProviderInfo,
    ResolveDirective,
)
from aioinject.scope import BaseScope, CurrentScope


if TYPE_CHECKING:
    from aioinject.container import Extensions, Registry


__all__ = ["CompilationParams", "compile_fn"]


@dataclasses.dataclass
class CompilationParams:
    root: AnyNode
    nodes: Sequence[AnyNode]
    scopes: type[BaseScope]


BODY = """
{async}def factory(scopes: "Mapping[BaseScope, Context]", current_scope: "BaseScope") -> "T":
{body}
    return {return_var_name}"""
PREPARE_SCOPE_CACHE = (
    "try:\n"
    "    {scope_name}_cache = scopes[{scope_name}].cache\n"
    "except KeyError as err:\n"
    '    err_msg = f"Requested scope {scope} not found, current scope is {{current_scope}}"\n'
    "    raise ScopeNotFoundError(err_msg)\n"
)

PREPARE_SCOPE_EXIT_STACK = (
    "{scope_name}_exit_stack = scopes[{scope_name}].exit_stack\n"
)
CHECK_CACHE = "if ({dependency}_instance := {scope_name}_cache.get({dependency}_type, NotInCache)) is NotInCache:\n"
CHECK_CACHE_STRICT = (
    "{dependency}_instance = {scope_name}_cache[{dependency}_type]\n"
)
CREATE_REGULAR_INSTANCE = (
    "{dependency}_instance = {await}{dependency}_provider.provide({kwargs})\n"
)
ACQUIRE_LOCK = "{async}with scopes[{scope_name}].lock:\n"
ACQUIRE_DOUBLE_CHECK_LOCK = (
    "{async}with scopes[{scope_name}].lock:\n"
    "    if ({dependency}_instance := {scope_name}_cache.get({dependency}_type, NotInCache)) is NotInCache:\n"
)
CREATE_CONTEXT_MANAGER_INSTANCE = "{dependency}_instance = {await}{scope_name}_exit_stack.{context_manager_method}({dependency}_provider.provide({kwargs}))\n"
STORE_CACHE = "{scope_name}_cache[{dependency}_type] = {dependency}_instance\n"
CALL_ON_RESOLVE_EXTENSION = (
    "for extension in registry.extensions.on_resolve:\n"
    "    await extension.on_resolve(context=scopes[{scope_name}], provider={dependency}_record, instance={dependency}_instance)\n"
)
CALL_SYNC_ON_RESOLVE_EXTENSION = (
    "for extension in registry.extensions.on_resolve_sync:\n"
    "     extension.on_resolve_sync(context=scopes[{scope_name}], provider={dependency}_record, instance={dependency}_instance)\n"
)
CALL_ON_RESOLVE_CONTEXT_EXTENSION = (
    "async with contextlib.AsyncExitStack() as cm:\n"
    "    for extension in registry.extensions.on_resolve_context:\n"
    "        await cm.enter_async_context(extension.on_resolve_context({dependency}_record))\n"
)


TCompilationDirective = TypeVar(
    "TCompilationDirective", bound=CompilationDirective
)


def get_directive(
    info: ProviderInfo[Any], directive: type[TCompilationDirective]
) -> TCompilationDirective | None:
    return next(
        (
            d
            for d in info.compilation_directives
            if isinstance(d, directive) and d.is_enabled
        ),
        None,
    )


def _compile_provider_node(  # noqa: C901
    node: ProviderNode,
    extensions: Extensions,
    *,
    is_async: bool,
) -> list[str]:
    parts: list[str] = []

    kwargs = generate_factory_kwargs(node.dependencies)
    indent = Indent(indent=1)

    provider = node.provider

    cache_directive = get_directive(provider.info, CacheDirective)
    resolve_directive = get_directive(provider.info, ResolveDirective)
    lock_directive = get_directive(provider.info, LockDirective)

    common_context = {
        "dependency": node.name,
        "kwargs": kwargs,
        "scope_name": f"{provider.info.scope.name}_scope",
    }

    if cache_directive:
        if cache_directive.optional:
            parts.append(indent.format(CHECK_CACHE.format_map(common_context)))
            indent.indent += 1
        else:  # pragma: no cover
            parts.append(
                indent.format(CHECK_CACHE_STRICT.format_map(common_context))
            )

    if lock_directive:
        part = (
            ACQUIRE_LOCK if not cache_directive else ACQUIRE_DOUBLE_CHECK_LOCK
        )
        context = common_context | {"async": "async " if is_async else ""}
        parts.append(indent.format(part).format_map(context))
        indent.indent += 1 if not cache_directive else 2

    if resolve_directive:
        context = common_context | {
            "context_manager_method": "enter_async_context"
            if resolve_directive.is_async
            else "enter_context",
            "await": "await " if resolve_directive.is_async else "",
        }

        if extensions.on_resolve_context and is_async:
            parts.append(
                indent.format(CALL_ON_RESOLVE_CONTEXT_EXTENSION).format_map(
                    context
                )
            )
            indent.indent += 1

        parts.append(
            indent.format(
                CREATE_REGULAR_INSTANCE
                if not resolve_directive.is_context_manager
                else CREATE_CONTEXT_MANAGER_INSTANCE
            ).format_map(context)
        )

    if cache_directive and cache_directive.optional:
        parts.append(indent.format(STORE_CACHE.format_map(common_context)))

    if resolve_directive:
        if is_async and extensions.on_resolve:
            parts.append(
                indent.format(
                    CALL_ON_RESOLVE_EXTENSION.format_map(common_context)
                )
            )

        if not is_async and extensions.on_resolve_sync:
            parts.append(
                indent.format(
                    CALL_SYNC_ON_RESOLVE_EXTENSION.format_map(common_context)
                )
            )

    return parts


def _compile_iterable_node(
    node: IterableNode,
) -> list[str]:
    deps = ", ".join(f"{dep.name}_instance" for dep in node.dependencies)
    template = f"    {node.name}_instance = [{deps}]\n"
    return [template]


def _compile_from_context_node(node: FromContextNode) -> list[str]:
    scope = (
        "current_scope"
        if isinstance(node.scope, CurrentScope)
        else f"{node.scope.name}_scope"
    )
    name = create_var_name(node)
    return [f"    {name}_instance = scopes[{scope}].cache[{name}_type]\n"]


def _compile_node(
    node: AnyNode,
    extensions: Extensions,
    *,
    is_async: bool,
) -> list[str]:
    match node:
        case ProviderNode():
            return _compile_provider_node(node, extensions, is_async=is_async)
        case IterableNode():
            return _compile_iterable_node(node)
        case FromContextNode():
            return _compile_from_context_node(node)
        case _:  # pragma: no cover
            typing.assert_never(node)  # type: ignore[unreachable]


def compile_fn(  # noqa: C901
    params: CompilationParams,
    registry: Registry,
    extensions: Extensions,
    *,
    is_async: bool,
) -> CompiledFn[Any]:
    namespace = {
        "NotInCache": object(),
        "ScopeNotFoundError": ScopeNotFoundError,
        "registry": registry,
        "contextlib": contextlib,
        **registry.type_context,
    }
    namespace.update(
        {f"{scope.name}_scope": scope for scope in registry.scopes}
    )

    for node in params.nodes:
        match node:
            case ProviderNode():
                provider = node.provider
                namespace[f"{create_var_name(node)}_provider"] = (
                    provider.provider
                )
                namespace[f"{create_var_name(node)}_record"] = provider
                namespace[f"{create_var_name(node)}_type"] = node.type_
            case FromContextNode():
                namespace[f"{create_var_name(node)}_type"] = node.type_
            case IterableNode():
                pass
            case _:  # pragma: no cover
                typing.assert_never(node)  # type: ignore[unreachable]

    parts = []

    used_scopes = set()
    for node in params.nodes:
        match node:
            case ProviderNode():
                used_scopes.add(node.provider.info.scope)
            case IterableNode() | FromContextNode():
                pass
            case _:  # pragma: no cover
                typing.assert_never(node)  # type: ignore[unreachable]

    for scope in used_scopes:
        indent = Indent(indent=1)
        parts.append(
            indent.format(
                PREPARE_SCOPE_CACHE.format_map(
                    {"scope_name": f"{scope.name}_scope", "scope": scope}
                )
            )
        )
        if any(
            directive.is_context_manager
            for node in params.nodes
            if isinstance(node, ProviderNode)
            and (
                directive := get_directive(
                    node.provider.info, ResolveDirective
                )
            )
        ):
            parts.append(
                indent.format(
                    PREPARE_SCOPE_EXIT_STACK.format_map(
                        {"scope_name": f"{scope.name}_scope"}
                    )
                )
            )

    for node in params.nodes:
        parts.extend(_compile_node(node, extensions, is_async=is_async))

    body = "".join(parts)
    return_var_name = create_var_name(params.root)
    module_src = BODY.format_map(
        {
            "body": body,
            "return_var_name": f"{create_var_name(params.root)}_instance",
            "async": "async " if is_async else "",
        }
    )
    source_filename = f"aioinject_{return_var_name}"
    linecache.cache[source_filename] = (
        len(module_src),
        None,
        module_src.splitlines(keepends=True),
        f"aioinject_{return_var_name}",
    )

    compiled = compile(module_src, source_filename, "exec")
    local_namespace: dict[str, Any] = {}
    exec(compiled, namespace, local_namespace)  # noqa: S102
    return local_namespace["factory"]



import contextlib
from collections.abc import AsyncIterator, Iterator
from typing import Any

from aioinject import Container, SyncContainer
from aioinject.extensions import LifespanExtension


class LifespanTestExtension(LifespanExtension):
    def __init__(self) -> None:
        self.open = False
        self.closed = False

    @contextlib.asynccontextmanager
    async def lifespan(
        self,
        _: Container,
    ) -> AsyncIterator[None]:
        self.open = True
        yield
        self.closed = True


async def test_lifespan_extension() -> None:
    extension = LifespanTestExtension()
    container = Container(extensions=[extension])
    assert not extension.closed
    assert not extension.open
    async with container:
        assert extension.open
        assert not extension.closed  # type: ignore[unreachable]
    assert extension.closed  # type: ignore[unreachable]


async def test_should_not_be_executed_by_sync_context() -> None:
    extension = LifespanTestExtension()
    container = SyncContainer(extensions=[extension])
    assert not extension.closed
    assert not extension.open
    with container:
        pass
    assert not extension.open
    assert not extension.closed


async def test_preserves_order() -> None:
    class ExtAsync:
        def __init__(self, elements: list[Any], element: Any) -> None:
            self.container = elements
            self.element = element

        @contextlib.asynccontextmanager
        async def lifespan(
            self,
            _: Container,
        ) -> AsyncIterator[None]:
            self.container.append(self.element)
            yield

    class ExtSync:
        def __init__(self, elements: list[Any], element: Any) -> None:
            self.container = elements
            self.element = element

        @contextlib.contextmanager
        def lifespan_sync(
            self,
            _: Container | SyncContainer,
        ) -> Iterator[None]:
            self.container.append(self.element)
            yield

    elements: list[str] = []
    container = Container(
        extensions=[ExtSync(elements, "sync"), ExtAsync(elements, "async")]
    )
    async with container:
        assert elements == ["sync", "async"]

    elements = []
    container = Container(
        extensions=[ExtAsync(elements, "async"), ExtSync(elements, "sync")]
    )
    async with container:
        assert elements == ["async", "sync"]



import uuid
from http import HTTPStatus

from django.test import Client


def test_function_route(django_test_client: Client) -> None:
    response = django_test_client.get("/dj/function")
    assert response.status_code == HTTPStatus.OK
    assert response.content == b"Number 42"


def test_function_route_with_parameter(django_test_client: Client) -> None:
    parameter = str(uuid.uuid4())
    response = django_test_client.get(f"/dj/function/{parameter}")
    assert response.status_code == HTTPStatus.OK
    assert response.content == f"Number {parameter} 42".encode()



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from aioinject import Container, Scoped
from aioinject.errors import CannotDetermineReturnTypeError


async def test_deferred_dependencies() -> None:
    if TYPE_CHECKING:
        from decimal import Decimal  # noqa: PLC0415

    def some_deferred_type() -> Decimal:
        from decimal import Decimal  # noqa: PLC0415

        return Decimal("1.0")

    class DoubledDecimal:
        def __init__(self, decimal: Decimal) -> None:
            self.decimal = decimal * 2

    container = Container()

    def register_decimal_scoped() -> None:
        from decimal import Decimal  # noqa: PLC0415

        container.register(Scoped(some_deferred_type, Decimal))

    register_decimal_scoped()
    container.register(Scoped(DoubledDecimal))
    async with container.context() as ctx:
        assert (await ctx.resolve(DoubledDecimal)).decimal == DoubledDecimal(
            some_deferred_type(),
        ).decimal


def test_provider_fn_deferred_dep_misuse() -> None:
    with pytest.raises(CannotDetermineReturnTypeError) as exc_info:
        from tests.container.mod_tests import (  # noqa: F401, PLC0415
            provider_fn_deferred_dep_misuse,
        )
    assert exc_info.match("Or it's type is not defined yet.")



from __future__ import annotations

import inspect
from collections.abc import Callable
from typing import TYPE_CHECKING

from fastapi.background import BackgroundTasks
from fastapi.requests import Request
from fastapi.websockets import WebSocket

import aioinject.scope
from aioinject.decorators import ContextParameter, base_inject
from aioinject.extensions import OnInitExtension


if TYPE_CHECKING:
    from starlette.types import ASGIApp, Receive, Scope, Send

    from aioinject import Container, SyncContainer

from aioinject import FromContext
from aioinject._types import (
    P,
    T,
    safe_issubclass,
    unwrap_annotated,
)


__all__ = ["AioInjectMiddleware", "inject"]


def inject(function: Callable[P, T]) -> Callable[P, T]:
    signature = inspect.signature(function)
    existing_parameter = next(
        (
            param
            for param in signature.parameters.values()
            if safe_issubclass(
                unwrap_annotated(param.annotation).type, (Request, WebSocket)
            )
        ),
        None,
    )
    parameter_name = (
        existing_parameter.name if existing_parameter else "aioinject__request"
    )
    parameter_type = (
        unwrap_annotated(existing_parameter.annotation).type
        if existing_parameter
        else Request
    )

    return base_inject(
        function,
        context_parameters=(
            ContextParameter(
                name=parameter_name,
                type_=parameter_type,
                remove=existing_parameter is None,
            ),
            ContextParameter(
                name="aioinject__background_tasks", type_=BackgroundTasks
            ),
        ),
        context_getter=lambda args, kwargs: kwargs[  # noqa: ARG005
            parameter_name
        ].state.aioinject_context,
    )


class AioInjectMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        container: Container,
    ) -> None:
        self.app = app
        self.container = container

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        async with self.container.context() as context:
            if scope["type"] == "http":
                request = Request(scope=scope, receive=receive, send=send)
                request.state.aioinject_context = context
            if scope["type"] == "websocket":
                ws = WebSocket(scope=scope, receive=receive, send=send)
                ws.state.aioinject_context = context
            await self.app(scope, receive, send)


class FastAPIExtension(OnInitExtension):
    def on_init(
        self,
        container: Container | SyncContainer,
    ) -> None:
        container.register(
            FromContext(Request, scope=aioinject.scope.Scope.request)
        )
        container.register(
            FromContext(BackgroundTasks, scope=aioinject.scope.Scope.request)
        )



import dataclasses
from functools import partial

import pytest
import strawberry
from strawberry import Schema

import aioinject
from aioinject import Container, Injected
from aioinject.ext.strawberry import AioInjectExtension, inject
from tests.integrations.conftest import NumberService


@pytest.mark.parametrize("field_name", ["number", "numberWithInfo"])
async def test_schema_execute(
    schema: Schema, container: Container, field_name: str
) -> None:
    query = f"""
    query {{ {field_name} }}
    """

    result = await schema.execute(query=query, context_value={})
    assert not result.errors

    async with container.context() as ctx:
        number = await ctx.resolve(int)

    assert result.data == {field_name: number}


async def test_custom_context(container: Container) -> None:
    @dataclasses.dataclass
    class Context:
        aioinject_context: aioinject.Context | aioinject.SyncContext | None = (
            None
        )

    def setter(
        context: Context,
        aioinject_context: aioinject.Context | aioinject.SyncContext,
    ) -> None:
        context.aioinject_context = aioinject_context

    def getter(context: Context) -> aioinject.Context | aioinject.SyncContext:
        return context.aioinject_context  # type: ignore[return-value]

    custom_inject = partial(inject, context_getter=getter)

    @strawberry.type
    class Query:
        @strawberry.field
        @custom_inject
        async def number(self, service: Injected[NumberService]) -> int:
            return service.get_number()

    schema = Schema(
        query=Query,
        extensions=[AioInjectExtension(container, context_setter=setter)],
    )

    query = """
    query {
        number
    }
    """
    result = await schema.execute(query=query, context_value=Context())
    assert not result.errors

    async with container.context() as ctx:
        number = await ctx.resolve(int)
    assert result.data == {"number": number}



import dataclasses

from aioinject import Container, Object, Scoped, Singleton
from aioinject.testing import TestContainer


async def test_override() -> None:
    container = Container()
    container.register(Scoped(lambda: 0, interface=int))

    testcontainer = TestContainer(container)

    override = 42

    async with (
        container.context() as ctx,
        testcontainer.override(Object(override)),
    ):
        assert await ctx.resolve(int) is override

    async with container.context() as ctx:
        assert await ctx.resolve(int) == 0

    # Test sync override
    async with (
        container.context() as ctx,
    ):
        with testcontainer.override(Object(override)):
            assert await ctx.resolve(int) is override

    async with container.context() as ctx:
        assert await ctx.resolve(int) == 0


async def test_should_not_override_unrelated_providers() -> None:
    class A:
        pass

    @dataclasses.dataclass
    class B:
        a: A

    @dataclasses.dataclass
    class C:
        b: B

    container = Container()
    container.register(Singleton(A), Singleton(B), Singleton(C))
    testcontainer = TestContainer(container)

    a = await container.root.resolve(A)
    b = await container.root.resolve(B)
    c = await container.root.resolve(C)

    async with testcontainer.override(Singleton(B)):
        c_override = await container.root.resolve(C)
        assert c_override is not c
        assert c_override.b is not b
        assert c_override.b.a is a

        assert await container.root.resolve(A) is a
        assert await container.root.resolve(B) is not b

    assert await container.root.resolve(A) is a
    assert await container.root.resolve(B) is b
    assert await container.root.resolve(C) is c


async def test_override_scoped_with_singleton() -> None:
    class A:
        pass

    @dataclasses.dataclass
    class B:
        a: A

    container = Container()
    container.register(Singleton(A), Scoped(B))
    testcontainer = TestContainer(container)

    a = await container.root.resolve(A)

    async with testcontainer.override(Singleton(B)):
        b = await container.root.resolve(B)
        assert b.a is a
        assert b is await container.root.resolve(B)
        assert await container.root.resolve(A) is a

    assert await container.root.resolve(A) is a

    async with container.context() as ctx:
        b_scoped = await ctx.resolve(B)
        assert b_scoped is not b
        assert b.a is a


async def test_should_not_remove_unrelated_objects_from_cache() -> None:
    class A:
        pass

    class B:
        pass

    container = Container()
    container.register(Singleton(A), Singleton(B))
    testcontainer = TestContainer(container)

    async with testcontainer.override(Singleton(B)):
        a = await container.root.resolve(A)
        b = await container.root.resolve(B)

    assert await container.root.resolve(A) is a
    assert await container.root.resolve(B) is not b



import aioinject


container = aioinject.SyncContainer()
container.register(aioinject.Singleton(list))

with container.context() as ctx:
    object_1 = ctx.resolve(list)

with container.context() as ctx:
    object_2 = ctx.resolve(list)

assert object_1 is object_2



from typing import NewType

from aioinject import Container, Object, Scoped


class _Interface:
    pass


class _A(_Interface):
    pass


class _Dependant:
    def __init__(self, interface: _Interface) -> None:
        self.interface = interface


async def test_ok() -> None:
    container = Container()
    container.register(Scoped(_A, interface=_Interface))
    container.register(Scoped(_Dependant))

    async with container.context() as context:
        result = await context.resolve(_Dependant)
        assert isinstance(result, _Dependant)
        assert isinstance(result.interface, _A)


async def test_object() -> None:
    container = Container()
    container.register(Object(_A(), interface=_Interface))
    container.register(Scoped(_Dependant))

    async with container.context() as context:
        result = await context.resolve(_Dependant)
        assert isinstance(result, _Dependant)
        assert isinstance(result.interface, _A)


async def test_object_newtype() -> None:
    A = NewType("A", str)
    B = NewType("B", str)

    container = Container()
    container.register(Object("A", A))
    container.register(Object("B", B))

    async with container.context() as context:
        a = await context.resolve(A)
        b = await context.resolve(B)
        assert a == "A"
        assert b == "B"


async def test_should_be_able_to_resolve_type_directly() -> None:
    container = Container()
    container.register(Scoped(_A, interface=_Interface))
    container.register(Scoped(_A))

    async with container.context() as context:
        assert await context.resolve(_A) is await context.resolve(_Interface)

    async with container.context() as context:
        assert await context.resolve(_Interface) is await context.resolve(_A)



import asyncio
import contextlib
from collections.abc import AsyncIterator
from datetime import datetime
from typing import NewType

from aioinject import (
    Container,
    Object,
    Scoped,
    Singleton,
    Transient,
)


class SingletonClient:
    pass


class DBConnection:
    pass


@contextlib.asynccontextmanager
async def setup_db_connection() -> AsyncIterator[DBConnection]:
    yield DBConnection()


Now = NewType("Now", datetime)


class Service:
    def __init__(
        self,
        now_a: Now,
        now_b: Now,
        int_object: int,
        connection: DBConnection,
        client: SingletonClient,
    ) -> None:
        self._now_a = now_a
        self._now_b = now_b
        self._int = int_object
        self._connection = connection
        self._client = client


async def main() -> None:
    container = Container()
    container.register(
        Singleton(SingletonClient),
        Scoped(setup_db_connection),
        Object(42),
        Transient(lambda: datetime.now(), interface=Now),  # noqa: DTZ005
        Scoped(Service),
    )

    async with container, container.context() as context:
        await context.resolve(Service)


if __name__ == "__main__":
    asyncio.run(main())



from collections.abc import Sequence
from typing import Generic, TypeVar

from aioinject import Container, FromContext, Object, Scoped, Singleton
from aioinject.scope import CurrentScope
from aioinject.validation.rules import DEFAULT_RULES
from aioinject.validation.validate import validate_or_err


T = TypeVar("T")


class _Logger:
    pass


class _LoggerA(_Logger):
    pass


class _LoggerB(_Logger):
    pass


class _DBConnection:
    pass


class _Model:
    pass


class _Request:
    pass


class _Repository(Generic[T]):
    def __init__(
        self, model: type[T], connection: _DBConnection, req: _Request
    ) -> None:
        self._model = model
        self._connection = connection
        self._req = req


class Service:
    def __init__(
        self,
        repository: _Repository[_Model],
        loggers: Sequence[_Logger],
    ) -> None:
        self._repository = repository
        self._loggers = loggers


async def test_ok() -> None:
    container = Container()
    container.register(
        Singleton(_DBConnection),
        Scoped(_Repository),
        Scoped(Service),
        Object(_LoggerA(), interface=_Logger),
        Object(_LoggerB(), interface=_Logger),
        Object(_Model, interface=type[_Model]),  # type: ignore[arg-type]
        FromContext(_Request, scope=CurrentScope()),
    )

    validate_or_err(container, DEFAULT_RULES)

    async with container, container.context({_Request: _Request()}) as ctx:
        await ctx.resolve(Service)  # Assert that dependency could be resolved



import inspect
import typing
from collections.abc import Mapping
from typing import Any, NewType

from aioinject._types import T, is_generic_alias
from aioinject.extensions import ProviderExtension
from aioinject.extensions.providers import (
    CacheDirective,
    ProviderInfo,
    ResolveDirective,
)
from aioinject.providers import Provider
from aioinject.scope import BaseScope, Scope


class Object(Provider[T]):
    def __init__(self, obj: T, interface: type[T] | None = None) -> None:
        self.implementation = obj
        self.interface = interface

    def provide(
        self,
        kwargs: Mapping[str, Any],  # noqa: ARG002
    ) -> T:
        return self.implementation

    def __repr__(self) -> str:  # pragma: no cover
        return f"{self.__class__.__name__}({self.implementation=}, {self.interface=})"


class ObjectProviderExtension(ProviderExtension[Object[Any]]):
    def __init__(self, default_scope: BaseScope = Scope.lifetime) -> None:
        self.default_scope = default_scope

    def supports_provider(self, provider: Object[object]) -> bool:
        return isinstance(provider, Object)

    def extract(
        self,
        provider: Object[T],
        type_context: Mapping[str, Any],  # noqa: ARG002
    ) -> ProviderInfo[T]:
        actual_type = typing.cast(
            "type[T]",
            type[provider.implementation]
            if inspect.isclass(provider.implementation)
            else type(provider.implementation),
        )
        if is_generic_alias(provider.interface) or isinstance(
            provider.interface, NewType
        ):
            actual_type = provider.interface  # type: ignore[assignment]

        return ProviderInfo(
            interface=provider.interface or actual_type,
            type_=actual_type,
            dependencies=(),
            scope=self.default_scope,
            compilation_directives=(
                CacheDirective(),
                ResolveDirective(is_async=False, is_context_manager=False),
            ),
        )



import contextlib
from collections.abc import Iterator

from aioinject import Container, SyncContainer
from aioinject.extensions import LifespanSyncExtension


class LifespanTestExtension(LifespanSyncExtension):
    def __init__(self) -> None:
        self.open: bool = False
        self.closed: bool = False

    @contextlib.contextmanager
    def lifespan_sync(
        self,
        _: Container | SyncContainer,
    ) -> Iterator[None]:
        self.open = True
        yield
        self.closed = True


async def test_async() -> None:
    extension = LifespanTestExtension()
    container = Container(extensions=[extension])
    assert not extension.closed
    assert not extension.open
    async with container:
        assert extension.open
        assert not extension.closed  # type: ignore[unreachable]
    assert extension.closed  # type: ignore[unreachable]


async def test_sync() -> None:
    extension = LifespanTestExtension()
    container = SyncContainer(extensions=[extension])
    assert not extension.closed
    assert not extension.open
    with container:
        assert extension.open
        assert not extension.closed  # type: ignore[unreachable]
    assert extension.closed  # type: ignore[unreachable]



from __future__ import annotations

import collections
import contextlib
import functools
import inspect
import sys
import types
import typing
from collections.abc import Awaitable, Callable, Iterator, Mapping, Sequence
from inspect import isclass
from types import GenericAlias
from typing import (
    TYPE_CHECKING,
    Any,
    NamedTuple,
    ParamSpec,
    TypeAlias,
    TypeGuard,
    TypeVar,
)

from typing_extensions import Self

from aioinject.errors import CannotDetermineReturnTypeError
from aioinject.scope import BaseScope


if TYPE_CHECKING:
    from aioinject.context import Context, SyncContext

T = TypeVar("T")
P = ParamSpec("P")
T_co = TypeVar("T_co", covariant=True)

FactoryResult: TypeAlias = (
    T
    | collections.abc.Awaitable[T]
    | collections.abc.Coroutine[Any, Any, T]
    | collections.abc.Iterator[T]
    | collections.abc.AsyncIterator[T]
)
FactoryType: TypeAlias = type[T] | Callable[..., FactoryResult[T]]

_GENERATORS = {
    collections.abc.Generator,
    collections.abc.Iterator,
}
_ASYNC_GENERATORS = {
    collections.abc.AsyncGenerator,
    collections.abc.AsyncIterator,
}

ExecutionContext = dict[BaseScope, "Context | SyncContext"]

CompiledFn = Callable[[ExecutionContext, BaseScope], Awaitable[T_co]]
SyncCompiledFn = Callable[[ExecutionContext, BaseScope], T_co]


class UnwrappedAnnotation(NamedTuple):
    type: type[object]
    args: Sequence[object]


def get_return_annotation(
    ret_annotation: str,
    context: dict[str, Any],
) -> type[Any]:
    return eval(ret_annotation, context)  # noqa: S307


def _get_function_namespace(fn: Callable[..., Any]) -> dict[str, Any]:
    return getattr(sys.modules.get(fn.__module__, None), "__dict__", {})


def _guess_return_type(  # noqa: C901
    factory: FactoryType[T],
    type_context: Mapping[str, type[object]],
) -> type[T]:
    if isinstance(factory, functools.partial):
        factory = factory.func
    unwrapped = inspect.unwrap(factory)

    origin = typing.get_origin(factory)
    is_generic = origin and inspect.isclass(origin)
    if isclass(factory) or is_generic:
        return typing.cast("type[T]", factory)

    try:
        return_type = typing.get_type_hints(
            unwrapped, include_extras=True, localns=type_context
        )["return"]
    except KeyError as e:
        msg = f"Factory {factory.__qualname__} does not specify return type."
        raise CannotDetermineReturnTypeError(msg) from e
    except NameError:
        # handle future annotations.
        # functions might have dependecies in them
        # and we don't have the container context here so
        # we can't call _get_type_hints
        ret_annotation = unwrapped.__annotations__["return"]

        try:
            return_type = get_return_annotation(
                ret_annotation,
                context=_get_function_namespace(unwrapped),
            )
        except NameError as e:
            msg = f"Factory {factory.__qualname__} does not specify return type. Or it's type is not defined yet."
            raise CannotDetermineReturnTypeError(msg) from e
    if origin := typing.get_origin(return_type):
        args = typing.get_args(return_type)

        is_async_gen = (
            origin in _ASYNC_GENERATORS
            and inspect.isasyncgenfunction(unwrapped)
        )
        is_sync_gen = origin in _GENERATORS and inspect.isgeneratorfunction(
            unwrapped,
        )
        if is_async_gen or is_sync_gen:
            return_type = args[0]

    # Classmethod returning `typing.Self`
    if return_type == Self and (
        self_cls := getattr(factory, "__self__", None)
    ):
        if not inspect.isclass(self_cls):
            return self_cls.__class__
        return self_cls

    return return_type


_sentinel = object()


@contextlib.contextmanager
def remove_annotation(
    annotations: dict[str, Any],
    name: str,
) -> Iterator[None]:
    annotation = annotations.pop(name, _sentinel)
    yield
    if annotation is not _sentinel:
        annotations[name] = annotation


def unwrap_annotated(type_hint: Any) -> UnwrappedAnnotation:
    if typing.get_origin(type_hint) is not typing.Annotated:
        return UnwrappedAnnotation(type_hint, ())

    dep_type, *args = typing.get_args(type_hint)
    return UnwrappedAnnotation(dep_type, tuple(args))


def is_iterable_generic_collection(type_: Any) -> bool:
    if not (origin := typing.get_origin(type_)):
        return False

    is_collection = collections.abc.Iterable in inspect.getmro(
        origin
    ) or safe_issubclass(origin, collections.abc.Iterable)
    return bool(is_collection and typing.get_args(type_))


def is_generic_alias(type_: Any) -> TypeGuard[GenericAlias]:
    return isinstance(
        type_,
        types.GenericAlias | typing._GenericAlias,  # type: ignore[attr-defined] # noqa: SLF001
    ) and not is_iterable_generic_collection(type_)


def safe_issubclass(
    obj: type[object], typ: type[object] | tuple[type[object], ...]
) -> bool:
    try:
        return issubclass(obj, typ)
    except TypeError:
        return False


def get_generic_origin(generic: type[object]) -> type[object]:
    if is_generic_alias(generic):
        return typing.get_origin(generic)
    return generic



import pytest
from strawberry import Schema

from aioinject import Container
from aioinject.ext.strawberry import AioInjectExtension
from tests.integrations.strawberry.schema import Query


@pytest.fixture
async def schema(container: Container) -> Schema:
    return Schema(query=Query, extensions=[AioInjectExtension(container)])


@pytest.fixture
def anyio_backend() -> str:
    return "asyncio"



import pytest

from aioinject import (
    Container,
    Scoped,
    Singleton,
    SyncContainer,
    Transient,
)
from aioinject.validation.errors import ValidationError
from aioinject.validation.rules import (
    NoAsyncDependenciesInSyncContainerRule,
)
from aioinject.validation.validate import validate_or_err


async def _async_dep() -> int:
    return 42


@pytest.mark.parametrize("provider_cls", [Scoped, Singleton, Transient])
def test_sync_container_err(provider_cls: type[Scoped[int]]) -> None:
    container = SyncContainer()
    container.register(provider_cls(_async_dep))
    with pytest.raises(ValidationError) as err_info:
        validate_or_err(container, (NoAsyncDependenciesInSyncContainerRule(),))

    assert str(err_info.value) == (
        "\n  async-dependency:\n    Provider(int) is async"
    )


def test_async_container_ok() -> None:
    container = Container()
    container.register(Scoped(_async_dep))
    validate_or_err(container, (NoAsyncDependenciesInSyncContainerRule(),))



import asyncio

from aiogram import Bot, Dispatcher, Router
from aiogram.filters import Command
from aiogram.types import Message

from aioinject import Container, Injected, Object
from aioinject.ext.aiogram import AioInjectMiddleware, inject


router = Router()


@router.message(
    Command(commands=["start"]),
)
@inject
async def start(
    message: Message,
    value: Injected[int],
) -> None:
    await message.reply(f"Injected value is {value}")


async def main() -> None:
    container = Container()
    container.register(Object(42))

    middleware = AioInjectMiddleware(container)
    middleware.add_to_router(router)

    dispatcher = Dispatcher()
    dispatcher.include_router(router)

    async with (
        container,
        Bot(token="token-here") as bot,  # noqa: S106
    ):
        await dispatcher.start_polling(bot)


if __name__ == "__main__":
    asyncio.run(main())



import dependency_injector.containers
import dependency_injector.providers
import di
import di.executors
import dishka
import lagom
import punq  # type: ignore[import-untyped]
import rodi
import wireup
from di.dependent import Dependent

import aioinject
from aioinject import Scoped
from benchmark.dependencies import (
    RepositoryA,
    RepositoryB,
    ServiceA,
    ServiceB,
    Session,
    UseCase,
)
from benchmark.lib.bench import (
    BenchmarkCollection,
    BenchmarkContext,
    ProjectUrl,
)


COMMON_DEPENDENCIES = [
    Session,
    RepositoryA,
    RepositoryB,
    ServiceA,
    ServiceB,
    UseCase,
]

simple_benchmarks = BenchmarkCollection()


@simple_benchmarks.bench(name="aioinject")
async def benchmark_aioinject(context: BenchmarkContext) -> None:
    container = aioinject.SyncContainer()
    container.register(*(Scoped(svc) for svc in COMMON_DEPENDENCIES))

    with container.context() as ctx:
        ctx.resolve(UseCase)

    for _ in range(context.rounds):
        with context.round(), container.context() as ctx:
            ctx.resolve(UseCase)


@simple_benchmarks.bench(
    name="dishka",
    extras=(ProjectUrl("https://github.com/reagento/dishka"),),
)
async def benchmark_dishka(context: BenchmarkContext) -> None:
    provider = dishka.Provider(scope=dishka.Scope.REQUEST)
    for svc in COMMON_DEPENDENCIES:
        provider.provide(svc)

    container = dishka.make_container(provider)

    with container() as ctx:
        ctx.get(UseCase)

    for _ in range(context.rounds):
        with context.round(), container() as ctx:
            ctx.get(UseCase)


@simple_benchmarks.bench(name="python")
async def bench_python(context: BenchmarkContext) -> None:
    for _ in range(context.rounds):
        with context.round():
            session = Session()
            repo_a = RepositoryA(session=session)
            repo_b = RepositoryB(session=session)
            svc_a = ServiceA(repository=repo_a)
            svc_b = ServiceB(repository=repo_b)
            UseCase(service_a=svc_a, service_b=svc_b)


@simple_benchmarks.bench(
    name="rodi", extras=(ProjectUrl("https://github.com/Neoteroi/rodi"),)
)
async def benchmark_rodi(context: BenchmarkContext) -> None:
    container = rodi.Container()
    for svc in COMMON_DEPENDENCIES:
        container.add_scoped(svc)

    container.resolve(UseCase)

    for _ in range(context.rounds):
        with context.round():
            container.resolve(UseCase)


@simple_benchmarks.bench(
    name="adriangb/di", extras=(ProjectUrl("https://github.com/adriangb/di"),)
)
async def benchmark_di(context: BenchmarkContext) -> None:
    container = di.Container()
    solved = container.solve(
        Dependent(UseCase, scope="request"), scopes=["request"]
    )

    executor = di.executors.SyncExecutor()
    with container.enter_scope("request") as state:
        solved.execute_sync(executor=executor, state=state)

    for _ in range(context.rounds):
        with context.round(), container.enter_scope("request") as state:
            solved.execute_sync(executor=executor, state=state)


@simple_benchmarks.bench(
    name="dependency-injector",
    extras=(
        ProjectUrl("https://github.com/ets-labs/python-dependency-injector"),
    ),
)
async def benchmark_dependency_injector(context: BenchmarkContext) -> None:
    class Container(dependency_injector.containers.DeclarativeContainer):
        session = dependency_injector.providers.Factory(Session)
        repository_a = dependency_injector.providers.Factory(
            RepositoryA, session=session
        )
        repository_b = dependency_injector.providers.Factory(
            RepositoryB, session=session
        )
        service_a = dependency_injector.providers.Factory(
            ServiceA, repository=repository_a
        )
        service_b = dependency_injector.providers.Factory(
            ServiceB, repository=repository_b
        )
        use_case = dependency_injector.providers.Factory(
            UseCase, service_a=service_a, service_b=service_b
        )

    container = Container()

    container.use_case()

    for _ in range(context.rounds):
        with context.round():
            container.use_case()


@simple_benchmarks.bench(
    name="punq",
    max_iterations=5_000,
    extras=(ProjectUrl("https://github.com/bobthemighty/punq"),),
)
async def benchmark_punq(context: BenchmarkContext) -> None:
    container = punq.Container()
    for svc in COMMON_DEPENDENCIES:
        container.register(svc)
    container.resolve(UseCase)

    for _ in range(context.rounds):
        with context.round():
            container.resolve(UseCase)


@simple_benchmarks.bench(
    name="lagom", extras=(ProjectUrl("https://github.com/meadsteve/lagom"),)
)
async def benchmark_lagom(context: BenchmarkContext) -> None:
    container = lagom.Container()

    for _ in range(context.rounds):
        with context.round():
            container[UseCase]


@simple_benchmarks.bench(
    name="wireup", extras=(ProjectUrl("https://github.com/maldoinc/wireup"),)
)
async def benchmark_wireup(context: BenchmarkContext) -> None:
    container = wireup.create_sync_container(
        services=[
            wireup.service(svc, lifetime="scoped")
            for svc in COMMON_DEPENDENCIES
        ]
    )
    for _ in range(context.rounds):
        with context.round(), container.enter_scope() as ctx:
            ctx.get(UseCase)



from django.urls import path
from rest_framework.routers import DefaultRouter

from tests.integrations.django.app.drf.views import DRFView, DRFViewSet
from tests.integrations.django.app.test.views import (
    function_view,
    function_view_with_parameter,
)


router = DefaultRouter()
router.register(prefix="drf/viewset", viewset=DRFViewSet, basename="number")


urlpatterns = [
    path("dj/function", function_view),
    path("dj/function/<str:parameter>", function_view_with_parameter),
    path("drf/api-view", DRFView.as_view()),
    *router.urls,
]



from typing import Final, Generic, TypeVar

from aioinject import Object, Scoped, SyncContainer


T = TypeVar("T")


class Box(Generic[T]):
    def __init__(self, value: T) -> None:
        self.value: Final = value

    def __repr__(self) -> str:
        return f"Box({self.value!r})"


container = SyncContainer()
container.register(
    Scoped(Box),
    Object("string value"),
    Object(42),
)

with container, container.context() as context:
    int_box = context.resolve(Box[int])
    print(int_box)  # Box(42)

    str_box = context.resolve(Box[str])
    print(str_box)  # Box('string value')

    container.register(Object(Box("bound"), interface=Box[str]))

    box_str = context.resolve(Box[str])
    print(box_str)  # Box('bound')



import contextlib
from collections.abc import AsyncIterator

import uvicorn
from fastapi import FastAPI

from aioinject import Container, Injected, Object
from aioinject.ext.fastapi import AioInjectMiddleware, FastAPIExtension, inject


container = Container(
    extensions=[FastAPIExtension()],  # (1)!
)
container.register(Object(42))


@contextlib.asynccontextmanager
async def lifespan(_: FastAPI) -> AsyncIterator[None]:
    async with container:  # (2)!
        yield


def create_app() -> FastAPI:
    app = FastAPI(lifespan=lifespan)
    app.add_middleware(AioInjectMiddleware, container=container)  # (3)!

    @app.get("/")
    @inject
    async def root(number: Injected[int]) -> int:
        return number

    return app


if __name__ == "__main__":
    uvicorn.run("main:create_app", factory=True, reload=True)



from __future__ import annotations

from collections.abc import Sequence

from benchmark.lib.bench import BenchmarkResult, ProjectUrl


def time_to_microsecond(delta: float) -> str:
    return f"{delta * 1_000_000:.3f}μs"


def time_to_ms(delta: float) -> str:
    return f"{delta * 1_000:.3f}ms"


def print_results(results: Sequence[BenchmarkResult]) -> None:
    row_template = "{:25} {:10} {:10} {:10} {:10}"
    print(  # noqa: T201
        row_template.format(
            "Name",
            "iterations",
            "total",
            "mean",
            "median",
            # "p95",
            # "p99",
        ),
    )
    for result in results:
        print(row_template.format(*get_result_columns(result)))  # noqa: T201


def get_result_columns(result: BenchmarkResult) -> tuple[str, ...]:
    total = time_to_ms(result.total)
    if result.extrapolated:
        total = (
            f"{time_to_ms(result.mean * result.params.rounds)} (extrapolated)"
        )
    return (
        result.name,
        str(result.rounds),
        total,
        time_to_microsecond(result.mean),
        time_to_microsecond(result.median),
    )


def print_markdown_table(results: Sequence[BenchmarkResult]) -> None:
    for result in results:
        project_url = next(
            (e for e in result.benchmark.extras if isinstance(e, ProjectUrl)),
            None,
        )
        if project_url is None:
            continue

        result.name = f"[{result.name}]({project_url.url})"

    header = ("Name", "iterations", "total", "mean", "median")
    rows = [header, *(get_result_columns(result) for result in results)]
    cell_sizes = []
    for i in range(len(rows[0])):
        max_length = max(len(row[i]) for row in rows)
        cell_sizes.append(max_length)
    row_template = " | ".join(f"{{:{cell_size}}}" for cell_size in cell_sizes)
    row_template = f"| {row_template} |"

    formatted_header = row_template.format(*rows[0])
    print(formatted_header)  # noqa: T201
    underline = "|".join("-" * (cell_size + 2) for cell_size in cell_sizes)
    underline = f"|{underline}|"
    print(underline)  # noqa: T201
    for row in rows[1:]:
        print(row_template.format(*row))  # noqa: T201



import contextlib
from collections.abc import AsyncIterator

from aioinject import Container
from aioinject.providers.scoped import Singleton


async def test_should_close_singletons() -> None:
    shutdown = False

    @contextlib.asynccontextmanager
    async def dependency() -> AsyncIterator[int]:
        nonlocal shutdown

        yield 42
        shutdown = True

    container = Container()
    container.register(Singleton(dependency))
    async with container:
        for _ in range(2):
            async with container.context() as ctx:
                assert await ctx.resolve(int) == 42  # noqa: PLR2004

        assert shutdown is False
    assert shutdown is True



from collections.abc import Iterable, MutableSequence, Sequence

import pytest

from aioinject._types import is_iterable_generic_collection


@pytest.mark.parametrize(
    ("type_", "is_iterable"),
    [
        (str, False),
        (Iterable, False),
        (Iterable[str], True),
        (list[str], True),
        (Sequence[str], True),
        (MutableSequence[str], True),
    ],
)
def test_is_iterable_generic_collection(
    type_: object,
    is_iterable: bool,
) -> None:
    assert is_iterable_generic_collection(type_) is is_iterable



from django.http import HttpRequest, HttpResponse

from aioinject import Injected
from aioinject.ext.django import inject


@inject
def root_view(
    _: HttpRequest,
    dependency: Injected[int],
) -> HttpResponse:
    return HttpResponse(content=f"{dependency}")



import dataclasses
import functools
import inspect
import typing
from collections.abc import (
    AsyncIterator,
    Awaitable,
    Callable,
    Iterator,
    Sequence,
)
from contextlib import nullcontext
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Generic,
    NamedTuple,
    TypeAlias,
)

from typing_extensions import TypeVar

from aioinject import Context, SyncContext
from aioinject._types import (
    P,
    T,
    remove_annotation,
    safe_issubclass,
    unwrap_annotated,
)
from aioinject.extensions.providers import Dependency


@dataclasses.dataclass(slots=True)
class Inject:
    pass


# Used to comply with Liskov's substitution principle
INJECTED: Any = object()

_F = TypeVar("_F", bound=Callable[..., Any])

if TYPE_CHECKING:
    Injected: TypeAlias = Annotated[T, Inject]

else:

    class Injected(Generic[T]):
        def __class_getitem__(cls, item: object) -> object:
            return Annotated[item, Inject]


def clear_wrapper(wrapper: _F, args: list[str]) -> _F:
    signature = inspect.signature(wrapper)
    new_params = tuple(
        p for p in signature.parameters.values() if p.name not in args
    )
    wrapper.__signature__ = signature.replace(  # type: ignore[attr-defined]
        parameters=new_params,
    )
    for name in args:
        del wrapper.__annotations__[name]
    return wrapper


def _find_inject_marker_in_annotated_args(
    args: Sequence[Any],
) -> Inject | None:
    for arg in args:
        if safe_issubclass(arg, Inject):
            return Inject()

        if isinstance(arg, Inject):
            return arg  # pragma: no cover
    return None


def collect_dependencies(
    dependant: typing.Callable[..., object],
) -> typing.Iterable[Dependency[object]]:
    with remove_annotation(dependant.__annotations__, "return"):
        type_hints = typing.get_type_hints(dependant, include_extras=True)

    for name, hint in type_hints.items():
        dep_type, args = unwrap_annotated(hint)
        inject_marker = _find_inject_marker_in_annotated_args(args)
        if inject_marker is None:
            continue

        yield Dependency(
            name=name,
            type_=dep_type,
        )


def add_parameters_to_signature(
    func: Callable[P, T],
    parameters: dict[str, type[object]],
) -> Callable[P, T]:
    signature = inspect.signature(func)

    existing_parameters = [
        param
        for param in signature.parameters.values()
        if param.name not in parameters
    ]
    kwargs_parameter = next(
        (
            param
            for param in existing_parameters
            if param.kind == inspect.Parameter.VAR_KEYWORD
        ),
        None,
    )
    if kwargs_parameter:
        existing_parameters.remove(kwargs_parameter)

    params = [
        *existing_parameters,
        *(
            inspect.Parameter(
                name=name,
                annotation=annotation,
                kind=inspect.Parameter.KEYWORD_ONLY,
            )
            for name, annotation in parameters.items()
        ),
    ]
    # Push **kwargs parameter to end of the function
    if kwargs_parameter is not None:
        params.append(kwargs_parameter)

    func.__signature__ = signature.replace(parameters=params)  # type: ignore[attr-defined]
    for name, annotation in parameters.items():
        func.__annotations__[name] = annotation
    return func


@dataclasses.dataclass(slots=True, kw_only=True)
class ContextParameter:
    type_: type[object]
    name: str
    remove: bool = True


ContextGetter = Callable[[tuple[Any, ...], dict[str, Any]], T]


def _add_context(
    context: Context | SyncContext,
    context_parameters: Sequence[ContextParameter],
    kwargs: dict[str, object],
) -> None:
    context.add_context(
        {
            parameter.type_: kwargs.pop(parameter.name)
            if parameter.remove
            else kwargs[parameter.name]
            for parameter in context_parameters
        }
    )


def _async_wrapper_factory(
    function: Callable[P, Awaitable[T]],
    context_parameters: Sequence[ContextParameter],
    context_getter: ContextGetter[Context],
    dependencies: Sequence[Dependency[object]],
    *,
    enter_context: bool,
) -> Callable[P, Awaitable[T]]:
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        context = context_getter(args, kwargs)
        async with (
            nullcontext(context) if not enter_context else context.context()
        ) as context:
            _add_context(context, context_parameters, kwargs)

            for dependency in dependencies:
                kwargs[dependency.name] = await context.resolve(
                    dependency.type_
                )
            return await function(*args, **kwargs)

    return wrapper


def _async_generator_wrapper_factory(
    function: Callable[P, AsyncIterator[T]],
    context_parameters: Sequence[ContextParameter],
    context_getter: ContextGetter[Context],
    dependencies: Sequence[Dependency[object]],
    *,
    enter_context: bool,
) -> Callable[P, AsyncIterator[T]]:
    async def wrapper(
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncIterator[T]:
        context = context_getter(args, kwargs)
        async with (
            nullcontext(context) if not enter_context else context.context()
        ) as context:
            _add_context(context, context_parameters, kwargs)

            for dependency in dependencies:
                kwargs[dependency.name] = await context.resolve(
                    dependency.type_
                )
            async for result in function(*args, **kwargs):
                yield result

    return wrapper


def _sync_wrapper_factory(
    function: Callable[P, Awaitable[T]],
    context_parameters: Sequence[ContextParameter],
    context_getter: ContextGetter[SyncContext],
    dependencies: Sequence[Dependency[object]],
    *,
    enter_context: bool,
) -> Callable[P, object]:
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> object:
        context = context_getter(args, kwargs)
        with (
            nullcontext(context) if not enter_context else context.context()
        ) as context:
            _add_context(context, context_parameters, kwargs)

            for dependency in dependencies:
                kwargs[dependency.name] = context.resolve(dependency.type_)
            return function(*args, **kwargs)

    return wrapper


def _sync_generator_wrapper_factory(
    function: Callable[P, Iterator[T]],
    context_parameters: Sequence[ContextParameter],
    context_getter: ContextGetter[SyncContext],
    dependencies: Sequence[Dependency[object]],
    *,
    enter_context: bool,
) -> Callable[P, Iterator[T]]:
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Iterator[T]:
        context = context_getter(args, kwargs)
        with (
            nullcontext(context) if not enter_context else context.context()
        ) as context:
            _add_context(context, context_parameters, kwargs)

            for dependency in dependencies:
                kwargs[dependency.name] = context.resolve(dependency.type_)

            yield from function(*args, **kwargs)

    return wrapper


class WrapperCacheKey(NamedTuple):
    is_async: bool
    is_generator: bool


_WRAPPERS = {
    WrapperCacheKey(is_async=True, is_generator=False): _async_wrapper_factory,
    WrapperCacheKey(
        is_async=True, is_generator=True
    ): _async_generator_wrapper_factory,
    WrapperCacheKey(is_async=False, is_generator=False): _sync_wrapper_factory,
    WrapperCacheKey(
        is_async=False, is_generator=True
    ): _sync_generator_wrapper_factory,
}


def base_inject(
    function: Callable[P, T],
    context_parameters: Sequence[ContextParameter],
    context_getter: ContextGetter[SyncContext | Context],
    *,
    enter_context: bool = False,
) -> Callable[P, T]:
    dependencies = list(collect_dependencies(function))

    is_async = inspect.iscoroutinefunction(
        function
    ) or inspect.isasyncgenfunction(function)
    is_generator = inspect.isgeneratorfunction(
        function
    ) or inspect.isasyncgenfunction(function)

    wrapper = _WRAPPERS[  # type: ignore[operator]
        WrapperCacheKey(is_async=is_async, is_generator=is_generator)
    ](
        function=function,
        context_parameters=context_parameters,
        enter_context=enter_context,
        dependencies=dependencies,
        context_getter=context_getter,
    )
    wrapper = functools.update_wrapper(wrapper=wrapper, wrapped=function)
    wrapper = clear_wrapper(
        wrapper,
        args=[dep.name for dep in dependencies],
    )
    return add_parameters_to_signature(
        wrapper, {p.name: p.type_ for p in context_parameters}
    )



from __future__ import annotations

import pytest

from aioinject import Container, Scoped


class A:
    def __init__(self, b: B) -> None: ...


class B:
    def __init__(self, a: A) -> None: ...


def test_cyclic_dependency() -> None:
    container = Container()
    container.register(Scoped(A), Scoped(B))
    with pytest.raises(ValueError) as exc_info:  # noqa: PT011
        container.registry.compile(A, is_async=False)

    assert str(exc_info.value) == (
        "Could not resolve dependencies for type <class 'tests.container.test_registry.B'>\n"
        "  unresolved dependencies: [<class 'tests.container.test_registry.A'>]"
    )



from collections.abc import Mapping
from typing import Any, TypeVar

from pydantic_settings import BaseSettings

from aioinject import Provider, Scope, SyncContainer
from aioinject.extensions import ProviderExtension
from aioinject.extensions.providers import (
    CacheDirective,
    ProviderInfo,
    ResolveDirective,
)


TSettings = TypeVar("TSettings", bound=BaseSettings)


class SettingsProvider(Provider[TSettings]):
    def __init__(self, settings_cls: type[TSettings]) -> None:
        self.implementation = settings_cls

    def provide(
        self,
        kwargs: dict[str, Any],  # noqa: ARG002
    ) -> TSettings:
        return self.implementation()


class SettingsProviderExtension(
    ProviderExtension[SettingsProvider[TSettings]],
):
    def supports_provider(self, provider: object) -> bool:
        return isinstance(provider, SettingsProvider)

    def extract(
        self,
        provider: SettingsProvider[TSettings],
        type_context: Mapping[str, type[object]],  # noqa: ARG002
    ) -> ProviderInfo[TSettings]:
        return ProviderInfo(
            interface=provider.implementation,
            type_=provider.implementation,
            dependencies=(),
            scope=Scope.lifetime,
            compilation_directives=(
                ResolveDirective(is_async=False, is_context_manager=False),
                CacheDirective(),
            ),
        )


class MySettings(BaseSettings):
    env_value: int


def main() -> None:
    container = SyncContainer(extensions=[SettingsProviderExtension()])
    container.register(SettingsProvider(MySettings))

    with container, container.context() as ctx:
        print(ctx.resolve(MySettings))  # env_value=...


if __name__ == "__main__":
    main()



from aioinject import Container, Transient


class A:
    pass


async def test_identity() -> None:
    class B:
        def __init__(self, a: A, b: A) -> None:
            self.a = a
            self.b = b

    container = Container()
    container.register(Transient(A))
    container.register(Transient(B))

    async with container.context() as ctx:
        b = await ctx.resolve(B)
        assert b.a is not b.b


async def test_identity_different_instances() -> None:
    class B:
        def __init__(self, a: A) -> None:
            self.a = a

    class C:
        def __init__(self, a: A, b: B) -> None:
            self.a = a
            self.b = b

    container = Container()
    container.register(Transient(A))
    container.register(Transient(B))
    container.register(Transient(C))

    async with container.context() as ctx:
        instance = await ctx.resolve(C)
        assert instance.a is not instance.b.a



from collections.abc import Iterable

from aioinject import Container, Scoped


class _Interface:
    pass


class _A(_Interface):
    pass


class _B(_Interface):
    pass


async def test_ok() -> None:
    container = Container()
    container.register(Scoped(_A, interface=_Interface))
    container.register(Scoped(_B, interface=_Interface))

    async with container.context() as context:
        result = list(await context.resolve(Iterable[_Interface]))  # type: ignore[type-abstract]
        assert isinstance(result[0], _A)
        assert isinstance(result[1], _B)


async def test_iterable_and_regular_dependency() -> None:
    """Test if we can provide an iterable dependency and a regular one"""

    class Dependant:
        def __init__(
            self, iterable: list[_Interface], interface: _Interface
        ) -> None:
            self.iterable = iterable
            self.interface = interface

    container = Container()
    container.register(Scoped(_A, interface=_Interface))
    container.register(Scoped(_B, interface=_Interface))
    container.register(Scoped(Dependant))

    async with container.context() as context:
        instance = await context.resolve(Dependant)

        assert isinstance(instance.iterable[0], _A)
        assert isinstance(instance.iterable[1], _B)

        assert instance.iterable[1] is instance.interface



from collections.abc import AsyncIterator, Iterator
from typing import Annotated

import httpx
import pytest
from fastapi import BackgroundTasks, Depends, FastAPI
from fastapi.requests import Request
from fastapi.testclient import TestClient
from fastapi.websockets import WebSocket
from httpx import ASGITransport

import aioinject
from aioinject import Injected
from aioinject.ext.fastapi import AioInjectMiddleware, FastAPIExtension, inject


@inject
def dependency(number: Injected[int]) -> int:
    return number


@inject
async def async_dependency(number: Injected[int]) -> int:
    return number


@inject
def generator_dependency(number: Injected[int]) -> Iterator[int]:
    yield number


@inject
async def async_generator_dependency(
    number: Injected[int],
) -> AsyncIterator[int]:
    yield number


@pytest.fixture
def container(provided_value: int) -> aioinject.Container:
    container = aioinject.Container(extensions=[FastAPIExtension()])
    container.register(aioinject.Object(provided_value))
    return container


@pytest.fixture
def app(container: aioinject.Container) -> FastAPI:  # noqa: C901
    app_ = FastAPI()
    app_.add_middleware(AioInjectMiddleware, container=container)

    @app_.get("/async-route")
    @inject
    async def function_route(
        provided: Injected[int],
    ) -> dict[str, str | int]:
        return {"value": provided}

    @app_.get("/sync-depends")
    @inject
    async def route_with_depends(
        number: Annotated[int, Depends(dependency)],
    ) -> dict[str, str | int]:
        return {"value": number}

    @app_.get("/async-depends")
    @inject
    async def route_with_async_depends(
        number: Annotated[int, Depends(async_dependency)],
    ) -> dict[str, str | int]:
        return {"value": number}

    @app_.get("/gen-depends")
    @inject
    async def route_with_gen_depends(
        number: Annotated[int, Depends(generator_dependency)],
    ) -> dict[str, str | int]:
        return {"value": number}

    @app_.get("/asyncgen-depends")
    @inject
    async def route_with_async_gen_depends(
        number: Annotated[int, Depends(async_generator_dependency)],
    ) -> dict[str, str | int]:
        return {"value": number}

    @app_.websocket("/ws")
    @inject
    async def websocket_route(ws: WebSocket, provided: Injected[int]) -> None:
        await ws.accept()
        await ws.send_json({"value": provided})

    @app_.get("/context-injection")
    @inject
    async def context_injection_route(
        request: Injected[Request], background_tasks: Injected[BackgroundTasks]
    ) -> bool:
        return isinstance(request, Request) and isinstance(
            background_tasks, BackgroundTasks
        )

    return app_


@pytest.fixture
async def http_client(app: FastAPI) -> AsyncIterator[httpx.AsyncClient]:
    async with httpx.AsyncClient(
        transport=ASGITransport(app),
        base_url="http://test",
    ) as client:
        yield client


@pytest.fixture
async def starlette_http_client(app: FastAPI) -> AsyncIterator[TestClient]:
    with TestClient(app=app) as client:
        yield client



import uvicorn
from litestar import Litestar, get

from aioinject import Container, Injected, Object
from aioinject.ext.litestar import AioInjectPlugin, inject


container = Container()
container.register(Object(42))


@get("/")
@inject
async def function_route(
    number: Injected[int],
) -> int:
    return number


def create_app() -> Litestar:
    return Litestar(
        [function_route],
        plugins=[AioInjectPlugin(container=container)],
    )


if __name__ == "__main__":
    uvicorn.run("main:create_app", factory=True, reload=True)



import di
import di.executors
import dishka
import wireup
from di import bind_by_type
from di.dependent import Dependent

import aioinject
from aioinject import Scoped
from benchmark.dependencies import (
    RepositoryA,
    RepositoryB,
    ServiceA,
    ServiceB,
    Session,
    UseCase,
    create_session,
    create_session_cm,
)
from benchmark.lib.bench import (
    BenchmarkCollection,
    BenchmarkContext,
    ProjectUrl,
)


COMMON_DEPENDENCIES = [
    RepositoryA,
    RepositoryB,
    ServiceA,
    ServiceB,
    UseCase,
]


context_benchmarks = BenchmarkCollection()


@context_benchmarks.bench(name="aioinject")
async def benchmark_aioinject(context: BenchmarkContext) -> None:
    container = aioinject.Container()
    container.register(Scoped(create_session_cm))
    container.register(*(Scoped(svc) for svc in COMMON_DEPENDENCIES))

    async with container.context() as ctx:
        await ctx.resolve(UseCase)

    for _ in range(context.rounds):
        with context.round():
            async with container.context() as ctx:
                await ctx.resolve(UseCase)


@context_benchmarks.bench(
    name="dishka", extras=(ProjectUrl("https://github.com/reagento/dishka"),)
)
async def benchmark_dishka(context: BenchmarkContext) -> None:
    provider = dishka.Provider(scope=dishka.Scope.REQUEST)
    provider.provide(create_session)
    for svc in COMMON_DEPENDENCIES:
        provider.provide(svc)

    container = dishka.make_async_container(provider)

    async with container() as ctx:
        await ctx.get(UseCase)

    for _ in range(context.rounds):
        with context.round():
            async with container() as ctx:
                await ctx.get(UseCase)


@context_benchmarks.bench(name="python")
async def bench_python(context: BenchmarkContext) -> None:
    for _ in range(context.rounds):
        with context.round():
            async with create_session_cm() as session:
                repo_a = RepositoryA(session=session)
                repo_b = RepositoryB(session=session)
                svc_a = ServiceA(repository=repo_a)
                svc_b = ServiceB(repository=repo_b)
                UseCase(service_a=svc_a, service_b=svc_b)


@context_benchmarks.bench(
    name="adriangb/di", extras=(ProjectUrl("https://github.com/adriangb/di"),)
)
async def benchmark_di(context: BenchmarkContext) -> None:
    container = di.Container()
    container.bind(
        bind_by_type(Dependent(create_session, scope="request"), Session)
    )
    solved = container.solve(
        Dependent(UseCase, scope="request"), scopes=["request"]
    )

    executor = di.executors.AsyncExecutor()
    async with container.enter_scope("request") as state:
        await solved.execute_async(executor=executor, state=state)

    for _ in range(context.rounds):
        with context.round():
            async with container.enter_scope("request") as state:
                await solved.execute_async(executor=executor, state=state)


@context_benchmarks.bench(
    name="wireup", extras=(ProjectUrl("https://github.com/maldoinc/wireup"),)
)
async def benchmark_wireup(context: BenchmarkContext) -> None:
    container = wireup.create_async_container(
        services=[
            wireup.service(create_session, lifetime="scoped"),
            *(
                wireup.service(svc, lifetime="scoped")
                for svc in COMMON_DEPENDENCIES
            ),
        ]
    )
    for _ in range(context.rounds):
        with context.round():
            async with container.enter_scope() as ctx:
                await ctx.get(UseCase)



from __future__ import annotations

from collections.abc import Iterator
from types import TracebackType
from typing import TYPE_CHECKING, Any

from typing_extensions import Self

from aioinject.context import ProviderRecord
from aioinject.extensions.providers import ProviderInfo


if TYPE_CHECKING:
    from aioinject import Container, Provider, SyncContainer
    from aioinject.container import Registry


def _dependant_providers(
    registry: Registry, root: ProviderRecord[object]
) -> Iterator[ProviderRecord[object]]:
    stack = [root]
    seen = []
    yield root
    while stack:
        provider = stack.pop()

        for provider_group in registry.providers.values():
            for dependant_provider in provider_group:
                if dependant_provider in seen:
                    continue
                seen.append(dependant_provider)
                dependant_types = [
                    dep.type_ for dep in dependant_provider.info.dependencies
                ]
                if provider.info.type_ in dependant_types:
                    yield dependant_provider
                stack.append(dependant_provider)


class _Override:
    def __init__(
        self, container: Container | SyncContainer, provider: Provider[object]
    ) -> None:
        self.container = container
        self.registry = container.registry
        self.provider = provider
        self.prev: list[ProviderRecord[Any]] | None = None
        self.prev_cache: dict[type[object], object] | None = None

        self.extension = self.registry.find_provider_extension(self.provider)
        self.info: ProviderInfo[Any] = self.extension.extract(
            self.provider, self.registry.type_context
        )

    async def __aenter__(self) -> Self:
        self._enter()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._exit()

    def __enter__(self) -> Self:
        self._enter()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._exit()

    def _enter(self) -> None:
        self.prev_cache = self.container.root.cache.copy()
        self.prev = self.registry.providers.get(self.info.interface)
        self._clear_provider(self.prev)

        self.registry.providers[self.info.interface] = [
            ProviderRecord(
                provider=self.provider,
                ext=self.extension,
                info=self.info,
            )
        ]

    def _exit(self) -> None:
        self._clear_provider(self.registry.providers[self.info.interface])
        if self.prev is not None:
            self.registry.providers[self.info.interface] = self.prev

        if self.prev_cache is not None:
            new_cache = self.container.root.cache
            self.container.root.cache = self.prev_cache
            self.container.root.cache.update(new_cache)

    def _clear_provider(
        self, providers: list[ProviderRecord[object]] | None
    ) -> None:
        if not providers:
            return  # pragma: no cover

        for provider in providers:
            for dependant in _dependant_providers(self.registry, provider):
                self.registry.compilation_cache.pop(
                    (dependant.info.type_, True), None
                )
                self.registry.compilation_cache.pop(
                    (dependant.info.type_, False), None
                )
                self.container.root.cache.pop(dependant.info.type_, None)


class TestContainer:
    __test__ = False  # pytest

    def __init__(self, container: Container | SyncContainer) -> None:
        self.container = container

    def override(self, provider: Provider[Any]) -> _Override:
        return _Override(self.container, provider)



from __future__ import annotations

import types
import typing
from collections.abc import Sequence
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from aioinject._compilation.resolve import AnyNode, BoundDependency


def create_var_name(node: AnyNode) -> str:
    return node.name


def make_dependency_name(type_: type[object]) -> str:
    args = typing.get_args(type_)
    type_id = id(type_ if not args else typing.get_origin(type_))
    if not args:
        return f"{type_.__name__}_{type_id}"

    args_str = "_".join(make_dependency_name(arg) for arg in args)
    if typing.get_origin(type_) == types.UnionType:
        return args_str

    return f"{type_.__name__}_{args_str}_{type_id}"


def generate_factory_kwargs(
    dependencies: Sequence[BoundDependency[object]],
) -> str:
    joined = ", ".join(
        f'"{dependency.name}": {dependency.variable_name}_instance'
        for dependency in dependencies
    )
    return "{" + joined + "}"



import functools
import random

from aioinject import Container, Scoped, Singleton
from aioinject._types import T


async def test_partial_kwargs() -> None:
    number_1 = random.randint(1, 1000)
    number_2 = random.randint(1, 1000)

    async def make_int_str() -> str:
        return f"{number_1}"

    async def fn(value: int, dependency: str) -> int:
        return value + int(dependency)

    container = Container()
    container.register(
        Scoped(functools.partial(fn, value=number_2)), Scoped(make_int_str)
    )
    async with container.context() as ctx:
        number = await ctx.resolve(int)
        assert number == number_1 + number_2


async def test_partial_args() -> None:
    async def fn(typ: type[T], dep: str) -> T:
        assert isinstance(dep, str)
        return typ()

    container = Container()
    container.register(
        Singleton(str), Scoped(functools.partial(fn, int), interface=int)
    )
    async with container.context() as ctx:
        number = await ctx.resolve(int)
        assert number == 0



from __future__ import annotations

import typing
from collections.abc import Mapping
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from typing import (
    TYPE_CHECKING,
    Any,
    Protocol,
    TypeVar,
    runtime_checkable,
)

from aioinject.context import ProviderRecord
from aioinject.extensions.providers import ProviderInfo


if TYPE_CHECKING:
    from aioinject import Container, Context, SyncContainer, SyncContext
    from aioinject._types import T

from aioinject.providers import Provider


_TProvider_contra = TypeVar(
    "_TProvider_contra", bound=Provider[Any], contravariant=True
)


@runtime_checkable
class LifespanExtension(Protocol):
    def lifespan(
        self,
        container: Container,
    ) -> AbstractAsyncContextManager[None]: ...


@runtime_checkable
class LifespanSyncExtension(Protocol):
    def lifespan_sync(
        self,
        container: Container | SyncContainer,
    ) -> AbstractContextManager[None]: ...


@runtime_checkable
class OnInitExtension(Protocol):
    def on_init(
        self,
        container: Container | SyncContainer,
    ) -> None: ...


@typing.runtime_checkable
class ProviderExtension(Protocol[_TProvider_contra]):
    def supports_provider(self, provider: _TProvider_contra) -> bool: ...

    def extract(
        self,
        provider: _TProvider_contra,
        type_context: Mapping[str, type[object]],
    ) -> ProviderInfo[Any]: ...


@runtime_checkable
class OnResolveExtension(Protocol):
    async def on_resolve(
        self,
        context: Context,
        provider: ProviderRecord[T],
        instance: T,
    ) -> None: ...


@runtime_checkable
class OnResolveSyncExtension(Protocol):
    def on_resolve_sync(
        self,
        context: SyncContext,
        provider: ProviderRecord[T],
        instance: T,
    ) -> None: ...


@runtime_checkable
class OnResolveContextExtension(Protocol):
    if TYPE_CHECKING:

        @property
        def enabled(self) -> bool: ...

    def on_resolve_context(
        self, provider: ProviderRecord[Any]
    ) -> AbstractAsyncContextManager[None]: ...


Extension = (
    ProviderExtension[Any]
    | OnInitExtension
    | LifespanExtension
    | LifespanSyncExtension
    | OnResolveExtension
    | OnResolveSyncExtension
    | OnResolveContextExtension
)



from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.viewsets import ViewSet

from aioinject import Injected
from aioinject.ext.django import inject


class DRFView(APIView):
    @inject
    def get(self, _: Request, number: Injected[int]) -> Response:
        return Response({"value": number})


class DRFViewSet(ViewSet):
    @inject
    def retrieve(
        self,
        _: Request,
        pk: int,
        number: Injected[int],
    ) -> Response:
        return Response({"id": pk, "value": number})



from __future__ import annotations

import collections
import dataclasses
import typing
from collections.abc import Iterator, Sequence
from types import GenericAlias
from typing import TYPE_CHECKING, Any, Generic

from aioinject._compilation.naming import make_dependency_name
from aioinject._types import (
    T,
    is_generic_alias,
    is_iterable_generic_collection,
)
from aioinject.context import ProviderRecord
from aioinject.errors import ProviderNotFoundError
from aioinject.providers.context import FromContext
from aioinject.providers.scoped import Transient
from aioinject.scope import BaseScope, CurrentScope


if TYPE_CHECKING:
    from aioinject.container import Registry
from aioinject.extensions.providers import Dependency


@dataclasses.dataclass(slots=True, kw_only=True)
class BoundDependency(Generic[T]):
    variable_name: str
    name: str
    type_: type[T]
    provider: ProviderRecord[T]


@dataclasses.dataclass(slots=True, kw_only=True)
class ProviderNode:
    name: str
    type_: GenericAlias | type[Any]
    provider: ProviderRecord[object]
    dependencies: tuple[BoundDependency[object], ...]

    def __hash__(self) -> int:
        return hash((self.name, False))


@dataclasses.dataclass(slots=True, kw_only=True)
class IterableNode:
    name: str
    type_: type[Any]
    inner_type: type[Any]
    dependencies: tuple[BoundDependency[object], ...]

    def __hash__(self) -> int:
        return hash((self.name, True))


@dataclasses.dataclass(slots=True, kw_only=True)
class FromContextNode:
    name: str
    type_: type[Any]
    scope: BaseScope | CurrentScope
    dependencies: tuple[BoundDependency[object], ...] = ()

    def __hash__(self) -> int:
        return hash((self.name, self.type_, self.scope))


AnyNode = ProviderNode | IterableNode | FromContextNode


def _get_orig_bases(
    type_: GenericAlias | type[Any],
) -> tuple[type, ...] | None:
    return getattr(type_, "__orig_bases__", None)


def generic_args_map(  # noqa: C901
    type_: GenericAlias | type[Any],
) -> dict[str, type[object]]:
    if is_generic_alias(type_):
        if not (parameters := getattr(type_.__origin__, "__parameters__", ())):
            return {}  # pragma: no cover

        params: dict[str, Any] = {
            param.__name__: param for param in parameters
        }
        return dict(zip(params, type_.__args__, strict=True))

    args_map = {}
    if orig_bases := _get_orig_bases(type_):
        # find the generic parent
        for base in orig_bases:
            if is_generic_alias(base):  # noqa: SIM102
                if params := {
                    param.__name__: param
                    for param in getattr(base.__origin__, "__parameters__", ())
                }:
                    args_map.update(
                        dict(zip(params, base.__args__, strict=True)),
                    )
    return args_map


def get_generic_arguments(type_: Any) -> list[typing.TypeVar] | None:
    """
    Returns generic arguments of given class, e.g. Class[T] would return [~T]
    """
    if is_generic_alias(type_):
        args = typing.get_args(type_)
        return [arg for arg in args if isinstance(arg, typing.TypeVar)]
    return None


def get_generic_parameter_map(
    provided_type: GenericAlias | type[Any],
    dependencies: Sequence[Dependency[Any]],
) -> dict[str, type[object]]:
    args_map = generic_args_map(provided_type)
    if not args_map:
        return {}

    result = {}
    for dependency in dependencies:
        inner_type = dependency.type_
        if dependency.type_.__name__ in args_map:
            result[dependency.name] = args_map[dependency.type_.__name__]

        if generic_arguments := get_generic_arguments(inner_type):
            # This is a generic type, we need to resolve the type arguments
            # and pass them to the provider.
            resolved_args = tuple(
                args_map[arg.__name__] for arg in generic_arguments
            )
            result[dependency.name] = inner_type[resolved_args]
    return result


def _resolve_provider_node_dependencies(
    type_: GenericAlias | type[Any],
    node_name: str,
    provider: ProviderRecord[object],
    registry: Registry,
) -> tuple[BoundDependency[Any], ...]:
    generic_args_map = get_generic_parameter_map(
        provided_type=type_,
        dependencies=provider.info.dependencies,
    )

    dependencies = []
    for provider_dependency in provider.info.dependencies:
        is_iterable = is_iterable_generic_collection(provider_dependency.type_)

        bound_dependency_type = generic_args_map.get(
            provider_dependency.name, provider_dependency.type_
        )
        dependency_type = (
            provider_dependency.type_ if is_iterable else bound_dependency_type
        )

        dependency_provider = registry.get_provider(
            typing.get_args(bound_dependency_type)[0]
            if is_iterable
            else dependency_type
        )
        dependency_args_map = get_generic_parameter_map(
            bound_dependency_type, dependency_provider.info.dependencies
        )
        resolved_type = (
            dependency_args_map.get(
                provider_dependency.name,
                dependency_type
                if is_generic_alias(dependency_type)
                else dependency_provider.info.type_,
            )
            if not is_iterable
            else dependency_type
        )
        variable_name = make_dependency_name(resolved_type)  # type: ignore[arg-type]

        if isinstance(dependency_provider.provider, Transient):
            variable_name = (
                f"{node_name}_{provider_dependency.name}_{variable_name}"
            )
        if isinstance(dependency_provider.provider, FromContext):
            scope_name = (
                provider.info.scope.name
                if isinstance(dependency_provider.provider.scope, CurrentScope)
                else dependency_provider.provider.scope.name
            )
            variable_name = f"{make_dependency_name(provider_dependency.type_)}_scope_{scope_name}"

        dependency = BoundDependency(
            variable_name=variable_name,
            name=provider_dependency.name,
            type_=resolved_type,  # type: ignore[arg-type]
            provider=dependency_provider,
        )
        dependencies.append(dependency)

    return tuple(dependencies)


def _resolve_node(
    type_: type[Any],
    name: str,
    registry: Registry,
    dependant: ProviderNode | None = None,
) -> AnyNode:
    try:
        provider = registry.get_provider(type_)
    except ProviderNotFoundError:
        if not is_iterable_generic_collection(type_):  # pragma: no cover
            raise

        inner_type = typing.get_args(type_)[0]
        providers = registry.get_providers(inner_type)

        return IterableNode(
            type_=type_,
            inner_type=inner_type,
            name=name,
            dependencies=tuple(
                BoundDependency(
                    name=make_dependency_name(provider.info.type_),
                    variable_name=make_dependency_name(provider.info.type_),
                    type_=provider.info.type_,
                    provider=provider,
                )
                for provider in providers
            ),
        )
    resolved_type: GenericAlias | type[Any] = (
        provider.info.type_ if type_ == provider.info.interface else type_
    )

    if isinstance(provider.provider, FromContext):
        scope = provider.provider.scope
        if dependant and isinstance(scope, CurrentScope):
            scope = dependant.provider.info.scope
            name = f"{make_dependency_name(type_)}_scope_{scope.name}"

        return FromContextNode(
            name=name,
            scope=scope,
            type_=type_,
        )

    return ProviderNode(
        type_=resolved_type,
        name=name,
        provider=provider,
        dependencies=_resolve_provider_node_dependencies(
            type_=resolved_type,
            node_name=name,
            provider=provider,
            registry=registry,
        ),
    )


def resolve_dependencies(  # noqa: C901
    root_type: type[Any],
    registry: Registry,
) -> Iterator[AnyNode]:
    stack = [
        _resolve_node(
            root_type, name=make_dependency_name(root_type), registry=registry
        )
    ]
    seen = set()
    while stack:
        node = stack.pop()
        if node in seen:
            continue

        seen.add(node)

        yield node

        match node:
            case ProviderNode():
                generic_args_map = get_generic_parameter_map(
                    node.type_, node.provider.info.dependencies
                )
                for dependency in node.dependencies:
                    orig_dependency = next(
                        dep
                        for dep in node.provider.info.dependencies
                        if dep.name == dependency.name
                    )
                    dependency_type = generic_args_map.get(
                        dependency.name, orig_dependency.type_
                    )
                    stack.append(
                        _resolve_node(
                            type_=dependency_type,
                            name=dependency.variable_name,
                            registry=registry,
                            dependant=node,
                        )
                    )

            case IterableNode():
                providers = registry.get_providers(node.inner_type)
                for provider in providers:
                    new_node = ProviderNode(
                        provider=provider,
                        name=make_dependency_name(provider.info.type_),
                        type_=provider.info.type_,
                        dependencies=_resolve_provider_node_dependencies(
                            type_=node.type_,
                            node_name=node.name,
                            provider=provider,
                            registry=registry,
                        ),
                    )
                    stack.append(new_node)
            case FromContextNode():
                pass
            case _:  # pragma: no cover
                typing.assert_never(node)  # type: ignore[unreachable]


def sort_nodes(nodes: Sequence[AnyNode]) -> Iterator[AnyNode]:
    postponed_nodes: dict[AnyNode, int] = collections.defaultdict(int)
    max_attempts = len(nodes)

    seen_types = set()
    queue: collections.deque[AnyNode] = collections.deque()
    for node in nodes:
        queue.appendleft(node)

    while queue:
        node = queue.pop()

        dependencies_satisfied = all(
            dep.type_ in seen_types for dep in node.dependencies
        )
        if not dependencies_satisfied:
            if postponed_nodes[node] >= max_attempts:
                msg = (
                    f"Could not resolve dependencies for type {node.type_}\n"
                    f"  unresolved dependencies: {[dep.type_ for dep in node.dependencies if dep.type_ not in seen_types]}"
                )
                raise ValueError(msg)
            postponed_nodes[node] += 1
            queue.insert(max(0, len(queue) - postponed_nodes[node]), node)
            continue

        yield node
        seen_types.add(node.type_)



import contextlib
from collections.abc import AsyncIterator
from typing import Annotated

from fastapi import Depends


class Session:
    pass


@contextlib.asynccontextmanager
async def create_session_cm() -> AsyncIterator[Session]:
    yield Session()


async def create_session() -> AsyncIterator[Session]:
    yield Session()


class RepositoryA:
    def __init__(
        self,
        session: Annotated[Session, Depends(create_session)],
    ) -> None:
        self._session = session


class RepositoryB:
    def __init__(
        self,
        session: Annotated[Session, Depends(create_session)],
    ) -> None:
        self._session = session


class ServiceA:
    def __init__(
        self,
        repository: Annotated[RepositoryA, Depends()],
    ) -> None:
        self._repository = repository


class ServiceB:
    def __init__(
        self,
        repository: Annotated[RepositoryB, Depends()],
    ) -> None:
        self._repository = repository


class UseCase:
    def __init__(
        self,
        service_a: Annotated[ServiceA, Depends()],
        service_b: Annotated[ServiceB, Depends()],
    ) -> None:
        self._service_a = service_a
        self._service_b = service_b

    async def execute(self) -> int:
        return 42



import contextlib
from collections.abc import AsyncIterator

from typing_extensions import Self


class PropagatedError(Exception):
    pass


class ExceptionPropagation:
    def __init__(self) -> None:
        self.exc: BaseException | None = None

    @contextlib.asynccontextmanager
    async def dependency(self) -> AsyncIterator[Self]:
        try:
            yield self
        except Exception as exc:
            self.exc = exc
            raise



from aioinject.decorators import add_parameters_to_signature


def test_add_parameters_to_signature_with_kwargs() -> None:
    def func(**kwargs: int) -> None:
        pass

    add_parameters_to_signature(func, parameters={"param": str})



from __future__ import annotations

from typing import Any, Protocol

from aioinject._types import FactoryResult, T_co


class Provider(Protocol[T_co]):
    implementation: Any

    def provide(self, kwargs: dict[str, Any]) -> FactoryResult[T_co]: ...



import contextlib
from collections.abc import AsyncIterator, Iterator
from typing import Any

import pytest

from aioinject import Container, Context, Injected, Scoped, SyncContainer
from aioinject.decorators import ContextParameter, base_inject


def context_getter(_: tuple[Any, ...], kwargs: dict[str, Any]) -> Any:
    return kwargs["context"]


class Dependency:
    def __init__(self) -> None:
        self.closed = False
        self.fn_done = False


@contextlib.asynccontextmanager
async def async_dependency() -> AsyncIterator[Dependency]:
    dependency = Dependency()
    try:
        yield dependency
    finally:
        dependency.closed = True


@pytest.fixture
async def async_container() -> Container:
    container = Container()
    container.register(Scoped(async_dependency))
    return container


@contextlib.contextmanager
def sync_dependency() -> Iterator[Dependency]:
    dependency = Dependency()
    try:
        yield dependency
    finally:
        dependency.closed = True


@pytest.fixture
def sync_container() -> SyncContainer:
    container = SyncContainer()
    container.register(Scoped(sync_dependency))
    return container


async def test_async(async_container: Container) -> None:
    async def fn(value: Injected[Dependency]) -> Dependency:
        value.fn_done = True
        return value

    fn = base_inject(
        fn,
        context_parameters=(
            ContextParameter(
                type_=Context,
                name="context",
            ),
        ),
        context_getter=context_getter,
    )

    async with async_container.context() as ctx:
        result = await fn(context=ctx)  # type: ignore[call-arg]
        assert isinstance(result, Dependency)
        assert result.fn_done
        assert not result.closed
    assert result.closed


async def test_async_iterator(async_container: Container) -> None:
    async def fn(value: Injected[Dependency]) -> AsyncIterator[Dependency]:
        yield value
        value.fn_done = True

    fn = base_inject(
        fn,
        context_parameters=(
            ContextParameter(
                type_=Context,
                name="context",
            ),
        ),
        context_getter=context_getter,
    )

    async with async_container.context() as ctx:
        iterator = aiter(fn(context=ctx))  # type: ignore[call-arg]
        result = await anext(iterator)
        assert isinstance(result, Dependency)
        assert not result.fn_done
        assert not result.closed

        with pytest.raises(StopAsyncIteration):
            await anext(iterator)
        assert result.fn_done
        assert not result.closed  # type: ignore[unreachable]

    assert result.closed  # type: ignore[unreachable]


def test_sync(sync_container: SyncContainer) -> None:
    def fn(value: Injected[Dependency]) -> Dependency:
        value.fn_done = True
        return value

    fn = base_inject(
        fn,
        context_parameters=(
            ContextParameter(
                type_=Context,
                name="context",
            ),
        ),
        context_getter=context_getter,
    )

    with sync_container.context() as ctx:
        result = fn(context=ctx)  # type: ignore[call-arg]
        assert isinstance(result, Dependency)
        assert result.fn_done
        assert not result.closed
    assert result.closed


def test_sync_iterator(sync_container: SyncContainer) -> None:
    def fn(value: Injected[Dependency]) -> Iterator[Dependency]:
        yield value
        value.fn_done = True

    fn = base_inject(
        fn,
        context_parameters=(
            ContextParameter(
                type_=Context,
                name="context",
            ),
        ),
        context_getter=context_getter,
    )

    with sync_container.context() as ctx:
        iterator = iter(fn(context=ctx))  # type: ignore[call-arg]
        result = next(iterator)
        assert isinstance(result, Dependency)
        assert not result.fn_done
        assert not result.closed

        with pytest.raises(StopIteration):
            next(iterator)
        assert result.fn_done
        assert not result.closed  # type: ignore[unreachable]

    assert result.closed  # type: ignore[unreachable]



import contextlib
from collections.abc import Iterator

import aioinject
from aioinject import Scoped, Singleton


@contextlib.contextmanager
def singleton_dependency() -> Iterator[int]:
    print("Singleton Startup")
    yield 42
    print("Singleton Shutdown")


@contextlib.contextmanager
def scoped_dependency(number: int) -> Iterator[str]:
    print("Scoped Startup")
    yield str(number)
    print("Scoped Shutdown")


container = aioinject.SyncContainer()
container.register(Singleton(singleton_dependency))
container.register(Scoped(scoped_dependency))

with container:  # noqa: SIM117
    with container.context() as ctx:
        value = ctx.resolve(str)  # Singleton Startup, Scoped Startup
        print(repr(value))  # '42'
    # Scoped Shutdown
# Singleton Shutdown



from aioinject import Container, SyncContainer
from aioinject.extensions import OnInitExtension


def test_on_mount_extension() -> None:
    class TestExtension(OnInitExtension):
        def __init__(self) -> None:
            self.mounted = False

        def on_init(self, _: Container | SyncContainer) -> None:
            self.mounted = True

    extension = TestExtension()
    assert not extension.mounted
    Container(extensions=[extension])
    assert extension.mounted



from __future__ import annotations

import collections
import typing
from collections.abc import Sequence
from types import TracebackType
from typing import Any, Final, Literal, TypeAlias

from typing_extensions import Self

from aioinject._compilation import (
    CompilationParams,
    compile_fn,
)
from aioinject._compilation.resolve import (
    resolve_dependencies,
    sort_nodes,
)
from aioinject._types import CompiledFn, SyncCompiledFn, T, get_generic_origin
from aioinject.context import Context, ProviderRecord, SyncContext
from aioinject.errors import ProviderNotFoundError
from aioinject.extensions import (
    Extension,
    LifespanExtension,
    LifespanSyncExtension,
    OnInitExtension,
    OnResolveContextExtension,
    OnResolveExtension,
    OnResolveSyncExtension,
    ProviderExtension,
)
from aioinject.extensions.providers import ProviderInfo
from aioinject.providers import Provider
from aioinject.providers.context import ContextProviderExtension
from aioinject.providers.object import ObjectProviderExtension
from aioinject.providers.scoped import ScopedProviderExtension
from aioinject.scope import BaseScope, Scope, next_scope


__all__ = ["Container", "Extensions", "Registry", "SyncContainer"]

DEFAULT_EXTENSIONS = (
    ScopedProviderExtension(),
    ObjectProviderExtension(),
    ContextProviderExtension(),
)


class Extensions:
    def __init__(
        self,
        extensions: Sequence[Extension],
        default_extensions: Sequence[Extension] = DEFAULT_EXTENSIONS,
    ) -> None:
        self._extensions: Final = tuple((*extensions, *default_extensions))  # noqa: C409

        self.providers = [
            e for e in self._extensions if isinstance(e, ProviderExtension)
        ]
        self.on_init = [
            e for e in self._extensions if isinstance(e, OnInitExtension)
        ]
        self.lifespan = [
            e
            for e in self._extensions
            if isinstance(e, LifespanExtension | LifespanSyncExtension)
        ]
        self.lifespan_sync = [
            e for e in self._extensions if isinstance(e, LifespanSyncExtension)
        ]
        self.on_resolve = [
            e for e in self._extensions if isinstance(e, OnResolveExtension)
        ]
        self.on_resolve_sync = [
            e
            for e in self._extensions
            if isinstance(e, OnResolveSyncExtension)
        ]
        self.on_resolve_context = [
            e
            for e in self._extensions
            if isinstance(e, OnResolveContextExtension) and e.enabled
        ]


RegistryCacheKey: TypeAlias = tuple[type[object], bool]


class Registry:
    def __init__(
        self, scopes: type[BaseScope], extensions: Extensions
    ) -> None:
        self.scopes = scopes
        self.extensions = extensions
        self.providers: dict[type[Any], list[ProviderRecord[Any]]] = (
            collections.defaultdict(list)
        )
        self.type_context: Final[dict[str, type[object]]] = {}
        self.compilation_cache: Final[
            dict[RegistryCacheKey, CompiledFn[Any]]
        ] = {}

    def register(self, *providers: Provider[Any]) -> None:
        for provider in providers:
            self._register_one(provider)

    def find_provider_extension(
        self, provider: Provider[Any]
    ) -> ProviderExtension[Any]:
        for ext in self.extensions.providers:
            if ext.supports_provider(provider):
                return ext

        err_msg = f"ProviderExtension for provider {provider!r} not found."
        raise ValueError(err_msg)

    def _register_one(self, provider: Provider[T]) -> None:
        ext = self.find_provider_extension(provider)
        if ext.supports_provider(provider):
            info: ProviderInfo[T] = ext.extract(
                provider, type_context=self.type_context
            )
            if any(
                provider.implementation
                == existing_provider.provider.implementation
                for existing_provider in self.providers.get(info.interface, [])
            ):
                msg = (
                    f"Provider for type {info.interface} with same "
                    f"implementation already registered"
                )
                raise ValueError(msg)

            self.providers[info.interface].append(
                ProviderRecord(
                    provider=provider,
                    info=info,
                    ext=ext,
                )
            )
            if class_name := info.type_.__name__:
                self.type_context[class_name] = get_generic_origin(info.type_)

    def get_providers(self, type_: type[T]) -> Sequence[ProviderRecord[T]]:
        if providers := self.providers.get(type_):
            return providers

        # Default to non-generic alias provider if there's one
        if (origin := typing.get_origin(type_)) and (
            providers := self.providers.get(origin)
        ):
            return providers

        err_msg = f"Providers for type {type_.__name__} not found"
        raise ProviderNotFoundError(err_msg)

    def get_provider(self, type_: type[T]) -> ProviderRecord[T]:
        return self.get_providers(type_)[-1]

    @typing.overload
    def compile(
        self,
        type_: type[T],
        *,
        is_async: Literal[True],
    ) -> CompiledFn[T]: ...

    @typing.overload
    def compile(
        self,
        type_: type[T],
        *,
        is_async: Literal[False],
    ) -> SyncCompiledFn[T]: ...

    def compile(
        self,
        type_: type[T],
        *,
        is_async: bool,
    ) -> CompiledFn[T] | SyncCompiledFn[T]:
        key = (type_, is_async)
        if key not in self.compilation_cache:
            nodes = list(resolve_dependencies(root_type=type_, registry=self))
            nodes.reverse()
            result = tuple(sort_nodes(nodes))

            self.compilation_cache[key] = compile_fn(
                CompilationParams(
                    root=result[-1],
                    nodes=result,
                    scopes=self.scopes,
                ),
                registry=self,
                extensions=self.extensions,
                is_async=is_async,
            )
        return self.compilation_cache[key]


def _run_on_init_extensions(container: Container | SyncContainer) -> None:
    for extension in container.extensions.on_init:
        extension.on_init(container)


class _BaseContainer:
    def __init__(
        self,
        extensions: Sequence[Extension],
        default_extensions: Sequence[Extension],
        scopes: type[BaseScope] = Scope,
    ) -> None:
        self.scopes: Final = scopes
        self.extensions = Extensions(
            extensions=extensions, default_extensions=default_extensions
        )
        self.registry = Registry(
            scopes=self.scopes, extensions=self.extensions
        )

    def register(self, *providers: Provider[Any]) -> None:
        self.registry.register(*providers)


class Container(_BaseContainer):
    def __init__(
        self,
        extensions: Sequence[Extension] = (),
        default_extensions: Sequence[Extension] = DEFAULT_EXTENSIONS,
        scopes: type[BaseScope] = Scope,
    ) -> None:
        super().__init__(
            extensions=extensions,
            default_extensions=default_extensions,
            scopes=scopes,
        )
        self._root: Context | None = None
        _run_on_init_extensions(self)

    def context(
        self, context: dict[type[object], object] | None = None
    ) -> Context:
        return self.root.context(context=context)

    @property
    def root(self) -> Context:
        if not self._root:
            self._root = Context(
                scope=next_scope(self.scopes, None),
                context={},
                container=self,
            )
        return self._root

    async def __aenter__(self) -> Self:
        for extension in self.extensions.lifespan:
            if isinstance(extension, LifespanExtension):
                await self.root.exit_stack.enter_async_context(
                    extension.lifespan(self)
                )
            if isinstance(extension, LifespanSyncExtension):
                self.root.exit_stack.enter_context(
                    extension.lifespan_sync(self)
                )

        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if self._root:
            await self._root.__aexit__(exc_type, exc_val, exc_tb)


class SyncContainer(_BaseContainer):
    def __init__(
        self,
        extensions: Sequence[Extension] = (),
        default_extensions: Sequence[Extension] = DEFAULT_EXTENSIONS,
        scopes: type[BaseScope] = Scope,
    ) -> None:
        super().__init__(
            extensions=extensions,
            default_extensions=default_extensions,
            scopes=scopes,
        )
        self._root: SyncContext | None = None
        _run_on_init_extensions(self)

    def __enter__(self) -> Self:
        for extension in self.extensions.lifespan_sync:
            self.root.exit_stack.enter_context(extension.lifespan_sync(self))

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if self._root:
            self._root.__exit__(exc_type, exc_val, exc_tb)

    @property
    def root(self) -> SyncContext:
        if not self._root:
            self._root = SyncContext(
                scope=next_scope(self.scopes, None),
                context={},
                container=self,
            )
        return self._root

    def context(
        self,
        context: dict[type[object], object] | None = None,
    ) -> SyncContext:
        return self.root.context(context=context)



from collections.abc import Sequence

from aioinject import Singleton, SyncContainer


class Logger:
    pass


class FileLogger(Logger):
    pass


class DatabaseLogger(Logger):
    pass


class StreamLogger(Logger):
    pass


container = SyncContainer()
container.register(
    Singleton(FileLogger, Logger),
    Singleton(DatabaseLogger, Logger),
    Singleton(StreamLogger, Logger),
)

with container, container.context() as context:
    loggers = context.resolve(Sequence[Logger])  # type: ignore[type-abstract]
    print(loggers)  # [<FileLogger>, <DatabaseLogger>, <StreamLogger>]



import contextlib
import dataclasses
from collections.abc import AsyncIterator
from datetime import datetime
from typing import Any, Final

import pytest

from aioinject import Container, Scoped, Singleton, SyncContainer, Transient
from aioinject.context import ProviderRecord
from aioinject.extensions import OnResolveContextExtension


def _now() -> datetime:
    return datetime.now()  # noqa: DTZ005


class _Singleton:
    pass


class _Repository:
    pass


@dataclasses.dataclass
class _Service:
    singleton: _Singleton
    now: datetime
    repository: _Repository


class _TestExtension(OnResolveContextExtension):
    def __init__(self, *, enabled: bool = True) -> None:
        self.enabled: Final = enabled
        self.provided: list[type[Any]] = []

    @contextlib.asynccontextmanager
    async def on_resolve_context(
        self, provider: ProviderRecord[Any]
    ) -> AsyncIterator[None]:
        yield
        self.provided.append(provider.info.interface)


@pytest.fixture
def extension() -> _TestExtension:
    return _TestExtension()


@pytest.fixture
def container(extension: OnResolveContextExtension) -> Container:
    container = Container(extensions=[extension])
    container.register(
        Transient(_now),
        Scoped(_Repository),
        Scoped(_Service),
        Singleton(_Singleton),
    )
    return container


async def test_executes_for_each_dependency_in_order(
    container: Container, extension: _TestExtension
) -> None:
    async with container.context() as ctx:
        await ctx.resolve(_Service)
        assert extension.provided == [
            _Singleton,
            datetime,
            _Repository,
            _Service,
        ]


async def test_does_not_execute_if_dependencies_are_cached(
    container: Container, extension: _TestExtension
) -> None:
    async with container.context() as ctx:
        await ctx.resolve(_Service)
        extension.provided.clear()

        await ctx.resolve(_Service)
        assert extension.provided == [datetime]


async def test_executes_once_for_singletons(
    container: Container, extension: _TestExtension
) -> None:
    async with container.context() as ctx:
        await ctx.resolve(_Service)
        assert _Singleton in extension.provided

    extension.provided.clear()

    async with container.context() as ctx:
        await ctx.resolve(_Service)
        assert _Singleton not in extension.provided


async def test_does_not_execute_if_disabled() -> None:
    extension = _TestExtension(enabled=False)
    container = Container(extensions=[extension])
    container.register(Scoped(int))

    async with container.context() as ctx:
        await ctx.resolve(int)
        assert not extension.provided


def test_async_extension_should_not_be_used_in_sync_container() -> None:
    container = SyncContainer(extensions=[_TestExtension()])
    container.register(Scoped(int))
    with container, container.context() as ctx:
        ctx.resolve(int)



from collections.abc import AsyncIterator
from typing import Any

import httpx
import pytest
from httpx import ASGITransport
from litestar import Controller, Litestar, Request, get, post
from litestar.config.app import AppConfig
from litestar.middleware import ASGIMiddleware
from litestar.plugins import InitPlugin
from litestar.types import ASGIApp, Receive, Send
from litestar.types import Scope as LitestarScope

import aioinject
from aioinject import Container, FromContext, Injected, Scope
from aioinject.ext.litestar import AioInjectPlugin, context_from_scope, inject
from tests.integrations.utils import ExceptionPropagation, PropagatedError


class Middleware(ASGIMiddleware):
    async def handle(
        self,
        scope: LitestarScope,
        receive: Receive,
        send: Send,
        next_app: ASGIApp,
    ) -> None:
        assert isinstance(context_from_scope(scope), aioinject.Context)
        await next_app(scope, receive, send)


class MiddlewarePlugin(InitPlugin):
    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        app_config.middleware.append(Middleware())
        return app_config


class LitestarController(Controller):
    @get("/exception-propagation")
    @inject
    async def exception_propagation(
        self, _: Injected[ExceptionPropagation]
    ) -> None:
        raise PropagatedError

    @get("/str")
    @inject
    async def str_route(self, value: Injected[str]) -> str:
        return value

    @post("/request-no-context")
    @inject
    async def request_context_1(self, request: Request[Any, Any, Any]) -> str:
        return (await request.body()).decode()

    @post("/request-context")
    @inject
    async def request_context_2(self, req: Injected[Request]) -> str:  # type: ignore[type-arg]
        return (await req.body()).decode()


@pytest.fixture
def app(container: Container) -> Litestar:
    container.register(FromContext(Request, scope=Scope.request))

    return Litestar(
        plugins=[AioInjectPlugin(container), MiddlewarePlugin()],
        route_handlers=[LitestarController],
        debug=True,
    )


@pytest.fixture
async def http_client(app: Litestar) -> AsyncIterator[httpx.AsyncClient]:
    async with (
        app.lifespan(),
        httpx.AsyncClient(
            transport=ASGITransport(app),  # type: ignore[arg-type]
            base_url="http://test",
        ) as client,
    ):
        yield client



from collections.abc import Mapping
from typing import Any

from aioinject._types import FactoryResult, T
from aioinject.extensions import ProviderExtension
from aioinject.extensions.providers import (
    ProviderInfo,
)
from aioinject.providers.abc import Provider
from aioinject.scope import BaseScope, CurrentScope


class FromContext(Provider[T]):
    def __init__(
        self,
        type_: type[T],
        scope: BaseScope | CurrentScope,
    ) -> None:
        self.implementation = type_
        self.scope = scope

    def provide(self, kwargs: dict[str, Any]) -> FactoryResult[T]:
        raise NotImplementedError


class ContextProviderExtension(ProviderExtension[FromContext[Any]]):
    def supports_provider(self, provider: Provider[Any]) -> bool:
        return isinstance(provider, FromContext)

    def extract(
        self,
        provider: FromContext[T],
        type_context: Mapping[str, type[object]],  # noqa: ARG002
    ) -> ProviderInfo[T]:
        return ProviderInfo(
            interface=provider.implementation,
            type_=provider.implementation,
            dependencies=(),
            scope=provider.scope,  # type: ignore[arg-type]
            compilation_directives=(),
        )



from __future__ import annotations

import dataclasses
import enum
import functools


@dataclasses.dataclass(slots=True)
class ScopeValue:
    key: str


class BaseScope(enum.Enum):
    def __init__(self, value: ScopeValue) -> None:
        self.key = value.key


class Scope(BaseScope):
    lifetime = ScopeValue("lifetime")
    request = ScopeValue("request")


@functools.cache
def next_scope(scopes: type[BaseScope], scope: BaseScope | None) -> BaseScope:
    members = list(scopes.__members__.values())
    if scope is None:
        return members[0]

    index = members.index(scope)
    return members[index + 1]


class CurrentScope:
    """Special marker to indicate resolving FromContext dependency from current scope."""



from __future__ import annotations

import contextlib
import inspect
import typing
from collections.abc import AsyncIterator, Callable
from typing import ParamSpec, TypeVar

from litestar import Litestar, Request
from litestar.config.app import AppConfig
from litestar.middleware import MiddlewareProtocol
from litestar.plugins import InitPluginProtocol
from litestar.types import ASGIApp, Receive, Scope, Send

from aioinject import Container, Context
from aioinject.decorators import ContextParameter, base_inject


__all__ = [
    "AioInjectMiddleware",
    "AioInjectPlugin",
    "context_from_scope",
    "inject",
]


_T = TypeVar("_T")
_P = ParamSpec("_P")

_CONTAINER_KEY = "__aioinject_container__"
_CONTEXT_KEY = "__aioinject_context__"

_REQUEST_PARAMETER_NAME = "request"


def inject(function: Callable[_P, _T]) -> Callable[_P, _T]:
    signature = inspect.signature(function)

    should_remove = (
        _REQUEST_PARAMETER_NAME not in signature.parameters
        or typing.get_origin(
            signature.parameters[_REQUEST_PARAMETER_NAME].annotation
        )
        is typing.Annotated
    )

    return base_inject(
        function=function,
        context_parameters=(
            ContextParameter(
                name=_REQUEST_PARAMETER_NAME,
                type_=Request,
                remove=should_remove,
            ),
        ),
        context_getter=lambda args, kwargs: kwargs[  # noqa: ARG005
            _REQUEST_PARAMETER_NAME
        ].scope[_CONTEXT_KEY],
    )


class AioInjectMiddleware(MiddlewareProtocol):
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        app: Litestar = scope["app"]
        container: Container = app.state[_CONTAINER_KEY]

        async with container.root.context() as context:
            scope[_CONTEXT_KEY] = context  # type: ignore[literal-required]
            await self.app(scope, receive, send)


async def _after_exception(exception: BaseException, scope: Scope) -> None:
    if _CONTEXT_KEY in scope:
        await scope[_CONTEXT_KEY].__aexit__(  # type: ignore[literal-required]
            type(exception),
            exception,
            exception.__traceback__,
        )


def context_from_scope(scope: Scope) -> Context:
    return scope[_CONTEXT_KEY]  # type: ignore[literal-required]


class AioInjectPlugin(InitPluginProtocol):
    def __init__(self, container: Container) -> None:
        self.container = container

    @contextlib.asynccontextmanager
    async def _lifespan(
        self,
        _: Litestar,
    ) -> AsyncIterator[None]:
        async with self.container:
            yield

    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        app_config.state[_CONTAINER_KEY] = self.container
        app_config.middleware.append(AioInjectMiddleware)
        app_config.lifespan.append(self._lifespan)
        app_config.after_exception.append(_after_exception)
        return app_config



import functools
import random

from aioinject import Scoped, SyncContainer
from aioinject.ext.django import SyncAioinjectMiddleware


@functools.cache
def create_container() -> SyncContainer:
    container = SyncContainer()
    container.register(
        Scoped(lambda: random.randint(1, 1000), interface=int),  # noqa: S311
    )
    return container


class DIMiddleware(SyncAioinjectMiddleware):
    container = create_container()



from __future__ import annotations

import dataclasses
import functools
from typing import Any, ParamSpec, TypeVar

import strawberry
import uvicorn
from starlette.requests import Request
from starlette.responses import Response
from starlette.websockets import WebSocket
from strawberry import Schema
from strawberry.asgi import GraphQL

import aioinject
from aioinject import Injected
from aioinject.ext.strawberry import AioInjectExtension
from aioinject.ext.strawberry import inject as base_inject


P = ParamSpec("P")
T = TypeVar("T")


def context_getter(
    context: Context,
) -> aioinject.Context | aioinject.SyncContext:
    return context.aioinject_context  # type: ignore[return-value]


def context_setter(
    context: Context,
    aioinject_context: aioinject.Context | aioinject.SyncContext,
) -> None:
    context.aioinject_context = aioinject_context


inject = functools.partial(base_inject, context_getter=context_getter)


@dataclasses.dataclass(slots=True, kw_only=True)
class Context:
    request: Request | WebSocket
    response: Response | WebSocket

    aioinject_context: aioinject.Context | aioinject.SyncContext | None = None


@strawberry.type
class Query:
    @strawberry.field
    @inject
    async def number(self, number: Injected[int]) -> int:
        return number


class MyGraphQL(GraphQL):
    async def get_context(
        self,
        request: Request | WebSocket,
        response: Response | WebSocket,
    ) -> Any:
        return Context(request=request, response=response)


def create_app() -> GraphQL[Any, Any]:
    container = aioinject.Container()
    container.register(aioinject.Object(42))

    schema = Schema(
        query=Query,
        extensions=[
            AioInjectExtension(
                container=container,
                context_setter=context_setter,
            ),
        ],
    )
    return MyGraphQL(schema=schema)


if __name__ == "__main__":
    uvicorn.run("main:create_app", factory=True, reload=True)



from __future__ import annotations

import functools
import inspect
import itertools
import typing
from collections.abc import Iterator, Mapping
from inspect import isclass

from aioinject._types import (
    FactoryType,
    remove_annotation,
    unwrap_annotated,
)
from aioinject.extensions.providers import Dependency


def _typevar_map(
    source: FactoryType[object],
) -> tuple[type, Mapping[object, object]]:
    origin = typing.get_origin(source)
    if not isclass(source) and not origin:
        return source, {}  # type: ignore[return-value]

    resolved_source = origin or source
    typevar_map: dict[object, object] = {}
    for base in (source, *getattr(source, "__orig_bases__", [])):
        origin = typing.get_origin(base)
        if not origin:
            continue

        params = getattr(origin, "__parameters__", ())
        args = typing.get_args(base)
        typevar_map |= dict(zip(params, args, strict=False))

    return resolved_source, typevar_map  # type: ignore[return-value]


def _get_ignored_partial_params(
    dependant: functools.partial[object],
) -> Iterator[str]:
    yield from dependant.keywords.keys()

    signature = inspect.signature(dependant.func)
    for param in itertools.islice(
        signature.parameters.values(), len(dependant.args)
    ):
        yield param.name


def collect_parameters(
    dependant: FactoryType[object],
    type_context: Mapping[str, type[object]],
) -> typing.Iterable[Dependency[object]]:
    """Collect parameter list from a function or class"""

    ignored_keywords: set[str] = set()
    if isinstance(dependant, functools.partial):
        ignored_keywords.update(_get_ignored_partial_params(dependant))
        dependant = dependant.func

    source, typevar_map = _typevar_map(source=dependant)

    if inspect.isclass(source):
        source = source.__init__

    with remove_annotation(getattr(source, "__annotations__", {}), "return"):
        type_hints = typing.get_type_hints(
            source, include_extras=True, localns=type_context
        )

    for name, hint in type_hints.items():
        if name in ignored_keywords:
            continue

        dep_type, _ = unwrap_annotated(hint)
        dep_type = typevar_map.get(dep_type, dep_type)  # type: ignore[assignment]
        yield Dependency(
            name=name,
            type_=dep_type,
        )
